##
## Copyright (c) 2006-2014 Red Hat, Inc. <http://www.redhat.com>
## This file is part of GlusterFS.
##
## This file is licensed to you under your choice of the GNU Lesser
## General Public License, version 3 or any later version (LGPLv3 or
## later), or the GNU General Public License, version 2 (GPLv2), in all
## cases as published by the Free Software Foundation.
##

import sys
import os
from ctypes import *

dl = CDLL(os.getenv("PATH_GLUSTERFS_GLUPY_MODULE", ""),RTLD_GLOBAL)


class call_frame_t (Structure):
        pass

class dev_t (Structure):
        pass


class dict_t (Structure):
        pass


class gf_dirent_t (Structure):
        pass


class iobref_t (Structure):
	pass


class iovec_t (Structure):
        pass


class list_head (Structure):
        pass

list_head._fields_ = [
                ("next", POINTER(list_head)),
                ("prev", POINTER(list_head))
        ]


class rwxperm_t (Structure):
        _fields_ = [
                ("read", c_uint8, 1),
                ("write", c_uint8, 1),
                ("execn", c_uint8, 1)
        ]


class statvfs_t (Structure):
        pass


class xlator_t (Structure):
        pass


class ia_prot_t (Structure):
        _fields_ = [
                ("suid", c_uint8, 1),
                ("sgid", c_uint8, 1),
                ("sticky", c_uint8, 1),
                ("owner", rwxperm_t),
                ("group", rwxperm_t),
                ("other", rwxperm_t)
        ]

# For checking file type.
(IA_INVAL, IA_IFREG, IA_IFDIR, IA_IFLNK, IA_IFBLK, IA_IFCHR, IA_IFIFO,
 IA_IFSOCK) = xrange(8)


class iatt_t (Structure):
        _fields_ = [
                ("ia_no", c_uint64),
                ("ia_gfid", c_ubyte * 16),
                ("ia_dev", c_uint64),
                ("ia_type", c_uint),
                ("ia_prot", ia_prot_t),
                ("ia_nlink", c_uint32),
                ("ia_uid", c_uint32),
                ("ia_gid", c_uint32),
                ("ia_rdev", c_uint64),
                ("ia_size", c_uint64),
                ("ia_blksize", c_uint32),
                ("ia_blocks", c_uint64),
                ("ia_atime", c_uint32 ),
                ("ia_atime_nsec", c_uint32),
                ("ia_mtime", c_uint32),
                ("ia_mtime_nsec", c_uint32),
                ("ia_ctime", c_uint32),
                ("ia_ctime_nsec", c_uint32)
        ]


class mem_pool (Structure):
        _fields_ = [
                ("list", list_head),
                ("hot_count", c_int),
                ("cold_count", c_int),
                ("lock", c_void_p),
                ("padded_sizeof_type", c_ulong),
                ("pool", c_void_p),
                ("pool_end", c_void_p),
                ("real_sizeof_type", c_int),
                ("alloc_count", c_uint64),
                ("pool_misses", c_uint64),
                ("max_alloc", c_int),
                ("curr_stdalloc", c_int),
                ("max_stdalloc", c_int),
                ("name", c_char_p),
                ("global_list", list_head)
        ]


class U_ctx_key_inode (Union):
        _fields_ = [
                ("key", c_uint64),
                ("xl_key", POINTER(xlator_t))
        ]


class U_ctx_value1 (Union):
        _fields_ = [
                ("value1", c_uint64),
                ("ptr1", c_void_p)
        ]


class U_ctx_value2 (Union):
        _fields_ = [
                ("value2", c_uint64),
                ("ptr2", c_void_p)
        ]

class inode_ctx (Structure):
        _anonymous_ = ("u_key","u_value1","u_value2",)
        _fields_ = [
                ("u_key", U_ctx_key_inode),
                ("u_value1", U_ctx_value1),
                ("u_value2", U_ctx_value2)
        ]

class inode_t (Structure):
        pass

class inode_table_t (Structure):
        _fields_ = [
                ("lock", c_void_p),
                ("hashsize", c_size_t),
                ("name", c_char_p),
                ("root", POINTER(inode_t)),
                ("xl", POINTER(xlator_t)),
                ("lru_limit", c_uint32),
                ("inode_hash", POINTER(list_head)),
                ("name_hash", POINTER(list_head)),
                ("active", list_head),
                ("active_size", c_uint32),
                ("lru", list_head),
                ("lru_size", c_uint32),
                ("purge", list_head),
                ("purge_size", c_uint32),
                ("inode_pool", POINTER(mem_pool)),
                ("dentry_pool", POINTER(mem_pool)),
                ("fd_mem_pool", POINTER(mem_pool))
        ]

inode_t._fields_ = [
                ("table", POINTER(inode_table_t)),
                ("gfid", c_ubyte * 16),
                ("lock", c_void_p),
                ("nlookup", c_uint64),
                ("fd_count", c_uint32),
                ("ref", c_uint32),
                ("ia_type", c_uint),
                ("fd_list", list_head),
                ("dentry_list", list_head),
                ("hashv", list_head),
                ("listv", list_head),
                ("ctx", POINTER(inode_ctx))
        ]



class U_ctx_key_fd (Union):
        _fields_ = [
                ("key", c_uint64),
                ("xl_key", c_void_p)
        ]

class fd_lk_ctx (Structure):
        _fields_ = [
                ("lk_list", list_head),
                ("ref", c_int),
                ("lock", c_void_p)
        ]

class fd_ctx (Structure):
        _anonymous_ = ("u_key","u_value1")
        _fields_ = [
                ("u_key", U_ctx_key_fd),
                ("u_value1", U_ctx_value1)
        ]

class fd_t (Structure):
        _fields_ = [
                ("pid", c_uint64),
                ("flags", c_int32),
                ("refcount", c_int32),
                ("inode_list", list_head),
                ("inode", POINTER(inode_t)),
                ("lock", c_void_p),
                ("ctx", POINTER(fd_ctx)),
                ("xl_count", c_int),
                ("lk_ctx", POINTER(fd_lk_ctx)),
                ("anonymous", c_uint)
        ]

class loc_t (Structure):
        _fields_ = [
                ("path", c_char_p),
                ("name", c_char_p),
                ("inode", POINTER(inode_t)),
                ("parent", POINTER(inode_t)),
                ("gfid", c_ubyte * 16),
                ("pargfid", c_ubyte * 16),
        ]



def _init_op (a_class, fop, cbk, wind, unwind):
        # Decorators, used by translators. We could pass the signatures as
        # parameters, but it's actually kind of nice to keep them around for
        # inspection.
        a_class.fop_type = apply(CFUNCTYPE,a_class.fop_sig)
        a_class.cbk_type = apply(CFUNCTYPE,a_class.cbk_sig)
        # Dispatch-function registration.
        fop.restype = None
        fop.argtypes = [ c_long, a_class.fop_type ]
        # Callback-function registration.
        cbk.restype = None
        cbk.argtypes = [ c_long, a_class.cbk_type ]
        # STACK_WIND function.
        wind.restype = None
        wind.argtypes = list(a_class.fop_sig[1:])
        # STACK_UNWIND function.
        unwind.restype = None
        unwind.argtypes = list(a_class.cbk_sig[1:])

class OpLookup:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(inode_t), POINTER(iatt_t),
                   POINTER(dict_t), POINTER(iatt_t))
_init_op (OpLookup, dl.set_lookup_fop, dl.set_lookup_cbk,
                    dl.wind_lookup,    dl.unwind_lookup)

class OpCreate:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), c_int, c_uint, c_uint, POINTER(fd_t),
                   POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(fd_t), POINTER(inode_t),
                   POINTER(iatt_t), POINTER(iatt_t), POINTER(iatt_t),
                   POINTER(dict_t))
_init_op (OpCreate, dl.set_create_fop, dl.set_create_cbk,
                    dl.wind_create,    dl.unwind_create)

class OpOpen:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), c_int, POINTER(fd_t), POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(fd_t), POINTER(dict_t))
_init_op (OpOpen, dl.set_open_fop, dl.set_open_cbk,
                  dl.wind_open,    dl.unwind_open)

class OpReadv:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(fd_t), c_size_t, c_long, c_uint32, POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(iovec_t), c_int, POINTER(iatt_t),
                   POINTER(iobref_t), POINTER(dict_t))
_init_op (OpReadv, dl.set_readv_fop, dl.set_readv_cbk,
                   dl.wind_readv,    dl.unwind_readv)
class OpWritev:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(fd_t), POINTER(iovec_t), c_int, c_long, c_uint32,
                   POINTER(iobref_t), POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(iatt_t), POINTER(iatt_t),
                   POINTER(dict_t))
_init_op (OpWritev, dl.set_writev_fop, dl.set_writev_cbk,
                    dl.wind_writev,    dl.unwind_writev)

class OpOpendir:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), POINTER(fd_t) ,POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(fd_t), POINTER(dict_t))
_init_op (OpOpendir, dl.set_opendir_fop, dl.set_opendir_cbk,
                     dl.wind_opendir,    dl.unwind_opendir)

class OpReaddir:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(fd_t), c_size_t, c_long, POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(gf_dirent_t), POINTER(dict_t))
_init_op (OpReaddir, dl.set_readdir_fop, dl.set_readdir_cbk,
                     dl.wind_readdir,    dl.unwind_readdir)

class OpReaddirp:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(fd_t), c_size_t, c_long, POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(gf_dirent_t), POINTER(dict_t))
_init_op (OpReaddirp, dl.set_readdirp_fop, dl.set_readdirp_cbk,
                      dl.wind_readdirp,    dl.unwind_readdirp)

class OpStat:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(iatt_t), POINTER(dict_t))
_init_op (OpStat, dl.set_stat_fop, dl.set_stat_cbk,
                  dl.wind_stat,    dl.unwind_stat)

class OpFstat:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(fd_t), POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(iatt_t), POINTER(dict_t))
_init_op (OpFstat, dl.set_fstat_fop, dl.set_fstat_cbk,
                   dl.wind_fstat,    dl.unwind_fstat)

class OpStatfs:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(statvfs_t), POINTER(dict_t))
_init_op (OpStatfs, dl.set_statfs_fop, dl.set_statfs_cbk,
                    dl.wind_statfs,    dl.unwind_statfs)


class OpSetxattr:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), POINTER(dict_t), c_int32,
                   POINTER (dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(dict_t))
_init_op (OpSetxattr, dl.set_setxattr_fop, dl.set_setxattr_cbk,
                      dl.wind_setxattr,    dl.unwind_setxattr)

class OpGetxattr:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), c_char_p, POINTER (dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(dict_t), POINTER(dict_t))
_init_op (OpGetxattr, dl.set_getxattr_fop, dl.set_getxattr_cbk,
                      dl.wind_getxattr,    dl.unwind_getxattr)

class OpFsetxattr:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(fd_t), POINTER(dict_t), c_int32,
                   POINTER (dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(dict_t))
_init_op (OpFsetxattr, dl.set_fsetxattr_fop, dl.set_fsetxattr_cbk,
                       dl.wind_fsetxattr,    dl.unwind_fsetxattr)

class OpFgetxattr:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(fd_t), c_char_p, POINTER (dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(dict_t), POINTER(dict_t))
_init_op (OpFgetxattr, dl.set_fgetxattr_fop, dl.set_fgetxattr_cbk,
                       dl.wind_fgetxattr,    dl.unwind_fgetxattr)

class OpRemovexattr:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), c_char_p, POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(dict_t))
_init_op (OpRemovexattr, dl.set_removexattr_fop, dl.set_removexattr_cbk,
                         dl.wind_removexattr,    dl.unwind_removexattr)


class OpFremovexattr:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(fd_t), c_char_p, POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(dict_t))
_init_op (OpFremovexattr, dl.set_fremovexattr_fop, dl.set_fremovexattr_cbk,
                          dl.wind_fremovexattr,    dl.unwind_fremovexattr)

class OpLink:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), POINTER(loc_t), POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(inode_t), POINTER(iatt_t),
                   POINTER(iatt_t), POINTER(iatt_t), POINTER(dict_t))
_init_op (OpLink, dl.set_link_fop, dl.set_link_cbk,
                  dl.wind_link,    dl.unwind_link)

class OpSymlink:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   c_char_p, POINTER(loc_t), c_uint, POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(inode_t), POINTER(iatt_t),
                   POINTER(iatt_t), POINTER(iatt_t), POINTER(dict_t))
_init_op (OpSymlink, dl.set_symlink_fop, dl.set_symlink_cbk,
                     dl.wind_symlink,    dl.unwind_symlink)

class OpUnlink:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), c_int, POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(iatt_t), POINTER(iatt_t),
                   POINTER(dict_t))
_init_op (OpUnlink, dl.set_unlink_fop, dl.set_unlink_cbk,
                    dl.wind_unlink,    dl.unwind_unlink)

class OpReadlink:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), c_size_t, POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, c_char_p, POINTER(iatt_t), POINTER(dict_t))
_init_op (OpReadlink, dl.set_readlink_fop, dl.set_readlink_cbk,
                      dl.wind_readlink,    dl.unwind_readlink)

class OpMkdir:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), c_uint, c_uint, POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(inode_t), POINTER(iatt_t),
                   POINTER(iatt_t), POINTER(iatt_t), POINTER(dict_t))
_init_op (OpMkdir, dl.set_mkdir_fop, dl.set_mkdir_cbk,
                   dl.wind_mkdir,    dl.unwind_mkdir)

class OpRmdir:
        fop_sig = (c_int, POINTER(call_frame_t), POINTER(xlator_t),
                   POINTER(loc_t), c_int, POINTER(dict_t))
        cbk_sig = (c_int, POINTER(call_frame_t), c_long, POINTER(xlator_t),
                   c_int, c_int, POINTER(iatt_t), POINTER(iatt_t),
                   POINTER(dict_t))
_init_op (OpRmdir, dl.set_rmdir_fop, dl.set_rmdir_cbk,
                   dl.wind_rmdir,    dl.unwind_rmdir)


class Translator:
        def __init__ (self, c_this):
                # This is only here to keep references to the stubs we create,
                # because ctypes doesn't and glupy.so can't because it doesn't
                # get a pointer to the actual Python object. It's a dictionary
                # instead of a list in case we ever allow changing fops/cbks
                # after initialization and need to look them up.
                self.stub_refs = {}
                funcs = dir(self.__class__)
                if "lookup_fop" in funcs:
                        @OpLookup.fop_type
                        def stub (frame, this, loc, xdata, s=self):
                                return s.lookup_fop (frame, this, loc, xdata)
                        self.stub_refs["lookup_fop"] = stub
                        dl.set_lookup_fop(c_this,stub)
                if "lookup_cbk" in funcs:
                        @OpLookup.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno, inode,
                                  buf, xdata, postparent, s=self):
                                return s.lookup_cbk(frame, cookie, this, op_ret,
                                                    op_errno, inode, buf, xdata,
                                                    postparent)
                        self.stub_refs["lookup_cbk"] = stub
                        dl.set_lookup_cbk(c_this,stub)
                if "create_fop" in funcs:
                        @OpCreate.fop_type
                        def stub (frame, this, loc, flags, mode, umask, fd,
                                  xdata, s=self):
                                return s.create_fop (frame, this, loc, flags,
                                                     mode, umask, fd, xdata)
                        self.stub_refs["create_fop"] = stub
                        dl.set_create_fop(c_this,stub)
                if "create_cbk" in funcs:
                        @OpCreate.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno, fd,
                                  inode, buf, preparent, postparent, xdata,
                                  s=self):
                                return s.create_cbk (frame, cookie, this,
                                                     op_ret, op_errno, fd,
                                                     inode, buf, preparent,
                                                     postparent, xdata)
                        self.stub_refs["create_cbk"] = stub
                        dl.set_create_cbk(c_this,stub)
                if "open_fop" in funcs:
                        @OpOpen.fop_type
                        def stub (frame, this, loc, flags, fd,
                                  xdata, s=self):
                                return s.open_fop (frame, this, loc, flags,
                                                   fd, xdata)
                        self.stub_refs["open_fop"] = stub
                        dl.set_open_fop(c_this,stub)
                if "open_cbk" in funcs:
                        @OpOpen.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno, fd,
                                  xdata, s=self):
                                return s.open_cbk (frame, cookie, this,
                                                   op_ret, op_errno, fd,
                                                   xdata)
                        self.stub_refs["open_cbk"] = stub
                        dl.set_open_cbk(c_this,stub)
                if "readv_fop" in funcs:
                        @OpReadv.fop_type
                        def stub (frame, this, fd, size, offset, flags,
                                  xdata, s=self):
                                return s.readv_fop (frame, this, fd, size,
                                                    offset, flags, xdata)
                        self.stub_refs["readv_fop"] = stub
                        dl.set_readv_fop(c_this,stub)
                if "readv_cbk" in funcs:
                        @OpReadv.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  vector, count, stbuf, iobref, xdata,
                                  s=self):
                                return s.readv_cbk (frame, cookie, this,
                                                    op_ret, op_errno, vector,
                                                    count, stbuf, iobref,
                                                    xdata)
                        self.stub_refs["readv_cbk"] = stub
                        dl.set_readv_cbk(c_this,stub)
                if "writev_fop" in funcs:
                        @OpWritev.fop_type
                        def stub (frame, this, fd, vector, count,
                                  offset, flags, iobref, xdata, s=self):
                                return s.writev_fop (frame, this, fd, vector,
                                                     count, offset, flags,
                                                     iobref, xdata)
                        self.stub_refs["writev_fop"] = stub
                        dl.set_writev_fop(c_this,stub)
                if "writev_cbk" in funcs:
                        @OpWritev.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  prebuf, postbuf, xdata, s=self):
                                return s.writev_cbk (frame, cookie, this,
                                                     op_ret, op_errno, prebuf,
                                                     postbuf, xdata)
                        self.stub_refs["writev_cbk"] = stub
                        dl.set_writev_cbk(c_this,stub)
                if "opendir_fop" in funcs:
                        @OpOpendir.fop_type
                        def stub (frame, this, loc, fd, xdata, s=self):
                                return s.opendir_fop (frame, this, loc, fd,
                                                      xdata)
                        self.stub_refs["opendir_fop"] = stub
                        dl.set_opendir_fop(c_this,stub)
                if "opendir_cbk" in funcs:
                        @OpOpendir.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno, fd,
                                  xdata, s=self):
                                return s.opendir_cbk(frame, cookie, this,
                                                     op_ret, op_errno, fd,
                                                     xdata)
                        self.stub_refs["opendir_cbk"] = stub
                        dl.set_opendir_cbk(c_this,stub)
                if "readdir_fop" in funcs:
                        @OpReaddir.fop_type
                        def stub (frame, this, fd, size, offset, xdata, s=self):
                                return s.readdir_fop (frame, this, fd, size,
                                                      offset, xdata)
                        self.stub_refs["readdir_fop"] = stub
                        dl.set_readdir_fop(c_this,stub)
                if "readdir_cbk" in funcs:
                        @OpReaddir.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  entries, xdata, s=self):
                                return s.readdir_cbk(frame, cookie, this,
                                                     op_ret, op_errno, entries,
                                                     xdata)
                        self.stub_refs["readdir_cbk"] = stub
                        dl.set_readdir_cbk(c_this,stub)
                if "readdirp_fop" in funcs:
                        @OpReaddirp.fop_type
                        def stub (frame, this, fd, size, offset, xdata, s=self):
                                return s.readdirp_fop (frame, this, fd, size,
                                                       offset, xdata)
                        self.stub_refs["readdirp_fop"] = stub
                        dl.set_readdirp_fop(c_this,stub)
                if "readdirp_cbk" in funcs:
                        @OpReaddirp.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  entries, xdata, s=self):
                                return s.readdirp_cbk (frame, cookie, this,
                                                       op_ret, op_errno,
                                                       entries, xdata)
                        self.stub_refs["readdirp_cbk"] = stub
                        dl.set_readdirp_cbk(c_this,stub)
                if "stat_fop" in funcs:
                        @OpStat.fop_type
                        def stub (frame, this, loc, xdata, s=self):
                                return s.stat_fop (frame, this, loc, xdata)
                        self.stub_refs["stat_fop"] = stub
                        dl.set_stat_fop(c_this,stub)
                if "stat_cbk" in funcs:
                        @OpStat.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno, buf,
                                  xdata, s=self):
                                return s.stat_cbk(frame, cookie, this, op_ret,
                                                  op_errno, buf, xdata)
                        self.stub_refs["stat_cbk"] = stub
                        dl.set_stat_cbk(c_this,stub)
                if "fstat_fop" in funcs:
                        @OpFstat.fop_type
                        def stub (frame, this, fd, xdata, s=self):
                                return s.fstat_fop (frame, this, fd, xdata)
                        self.stub_refs["fstat_fop"] = stub
                        dl.set_fstat_fop(c_this,stub)
                if "fstat_cbk" in funcs:
                        @OpFstat.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno, buf,
                                  xdata, s=self):
                                return s.fstat_cbk(frame, cookie, this, op_ret,
                                                   op_errno, buf, xdata)
                        self.stub_refs["fstat_cbk"] = stub
                        dl.set_fstat_cbk(c_this,stub)
                if "statfs_fop" in funcs:
                        @OpStatfs.fop_type
                        def stub (frame, this, loc, xdata, s=self):
                                return s.statfs_fop (frame, this, loc, xdata)
                        self.stub_refs["statfs_fop"] = stub
                        dl.set_statfs_fop(c_this,stub)
                if "statfs_cbk" in funcs:
                        @OpStatfs.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno, buf,
                                  xdata, s=self):
                                return s.statfs_cbk (frame, cookie, this,
                                                     op_ret, op_errno, buf,
                                                     xdata)
                        self.stub_refs["statfs_cbk"] = stub
                        dl.set_statfs_cbk(c_this,stub)
                if "setxattr_fop" in funcs:
                        @OpSetxattr.fop_type
                        def stub (frame, this, loc, dictionary, flags, xdata,
                                  s=self):
                                return s.setxattr_fop (frame, this, loc,
                                                       dictionary, flags,
                                                       xdata)
                        self.stub_refs["setxattr_fop"] = stub
                        dl.set_setxattr_fop(c_this,stub)
                if "setxattr_cbk" in funcs:
                        @OpSetxattr.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno, xdata,
                                  s=self):
                                return s.setxattr_cbk(frame, cookie, this,
                                                      op_ret, op_errno, xdata)
                        self.stub_refs["setxattr_cbk"] = stub
                        dl.set_setxattr_cbk(c_this,stub)
                if "getxattr_fop" in funcs:
                        @OpGetxattr.fop_type
                        def stub (frame, this, loc, name, xdata, s=self):
                                return s.getxattr_fop (frame, this, loc, name,
                                                       xdata)
                        self.stub_refs["getxattr_fop"] = stub
                        dl.set_getxattr_fop(c_this,stub)
                if "getxattr_cbk" in funcs:
                        @OpGetxattr.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  dictionary, xdata, s=self):
                                return s.getxattr_cbk(frame, cookie, this,
                                                      op_ret, op_errno,
                                                      dictionary, xdata)
                        self.stub_refs["getxattr_cbk"] = stub
                        dl.set_getxattr_cbk(c_this,stub)
                if "fsetxattr_fop" in funcs:
                        @OpFsetxattr.fop_type
                        def stub (frame, this, fd, dictionary, flags, xdata,
                                  s=self):
                                return s.fsetxattr_fop (frame, this, fd,
                                                        dictionary, flags,
                                                        xdata)
                        self.stub_refs["fsetxattr_fop"] = stub
                        dl.set_fsetxattr_fop(c_this,stub)
                if "fsetxattr_cbk" in funcs:
                        @OpFsetxattr.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno, xdata,
                                  s=self):
                                return s.fsetxattr_cbk(frame, cookie, this,
                                                       op_ret, op_errno, xdata)
                        self.stub_refs["fsetxattr_cbk"] = stub
                        dl.set_fsetxattr_cbk(c_this,stub)
                if "fgetxattr_fop" in funcs:
                        @OpFgetxattr.fop_type
                        def stub (frame, this, fd, name, xdata, s=self):
                                return s.fgetxattr_fop (frame, this, fd, name,
                                                        xdata)
                        self.stub_refs["fgetxattr_fop"] = stub
                        dl.set_fgetxattr_fop(c_this,stub)
                if "fgetxattr_cbk" in funcs:
                        @OpFgetxattr.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  dictionary, xdata, s=self):
                                return s.fgetxattr_cbk(frame, cookie, this,
                                                       op_ret, op_errno,
                                                       dictionary, xdata)
                        self.stub_refs["fgetxattr_cbk"] = stub
                        dl.set_fgetxattr_cbk(c_this,stub)
                if "removexattr_fop" in funcs:
                        @OpRemovexattr.fop_type
                        def stub (frame, this, loc, name, xdata, s=self):
                                return s.removexattr_fop (frame, this, loc,
                                                          name, xdata)
                        self.stub_refs["removexattr_fop"] = stub
                        dl.set_removexattr_fop(c_this,stub)
                if "removexattr_cbk" in funcs:
                        @OpRemovexattr.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  xdata, s=self):
                                return s.removexattr_cbk(frame, cookie, this,
                                                         op_ret, op_errno,
                                                         xdata)
                        self.stub_refs["removexattr_cbk"] = stub
                        dl.set_removexattr_cbk(c_this,stub)
                if "fremovexattr_fop" in funcs:
                        @OpFremovexattr.fop_type
                        def stub (frame, this, fd, name, xdata, s=self):
                                return s.fremovexattr_fop (frame, this, fd,
                                                           name, xdata)
                        self.stub_refs["fremovexattr_fop"] = stub
                        dl.set_fremovexattr_fop(c_this,stub)
                if "fremovexattr_cbk" in funcs:
                        @OpFremovexattr.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  xdata, s=self):
                                return s.fremovexattr_cbk(frame, cookie, this,
                                                          op_ret, op_errno,
                                                          xdata)
                        self.stub_refs["fremovexattr_cbk"] = stub
                        dl.set_fremovexattr_cbk(c_this,stub)
                if "link_fop" in funcs:
                        @OpLink.fop_type
                        def stub (frame, this, oldloc, newloc,
                                  xdata, s=self):
                                return s.link_fop (frame, this, oldloc,
                                                   newloc, xdata)
                        self.stub_refs["link_fop"] = stub
                        dl.set_link_fop(c_this,stub)
                if "link_cbk" in funcs:
                        @OpLink.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  inode, buf, preparent, postparent, xdata,
                                  s=self):
                                return s.link_cbk (frame, cookie, this,
                                                   op_ret, op_errno, inode,
                                                   buf, preparent,
                                                   postparent, xdata)
                        self.stub_refs["link_cbk"] = stub
                        dl.set_link_cbk(c_this,stub)
                if "symlink_fop" in funcs:
                        @OpSymlink.fop_type
                        def stub (frame, this, linkname, loc,
                                  umask, xdata, s=self):
                                return s.symlink_fop (frame, this, linkname,
                                                      loc, umask, xdata)
                        self.stub_refs["symlink_fop"] = stub
                        dl.set_symlink_fop(c_this,stub)
                if "symlink_cbk" in funcs:
                        @OpSymlink.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  inode, buf, preparent, postparent, xdata,
                                  s=self):
                                return s.symlink_cbk (frame, cookie, this,
                                                      op_ret, op_errno, inode,
                                                      buf, preparent,
                                                      postparent, xdata)
                        self.stub_refs["symlink_cbk"] = stub
                        dl.set_symlink_cbk(c_this,stub)
                if "unlink_fop" in funcs:
                        @OpUnlink.fop_type
                        def stub (frame, this, loc, xflags,
                                  xdata, s=self):
                                return s.unlink_fop (frame, this, loc,
                                                     xflags, xdata)
                        self.stub_refs["unlink_fop"] = stub
                        dl.set_unlink_fop(c_this,stub)
                if "unlink_cbk" in funcs:
                        @OpUnlink.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  preparent, postparent, xdata, s=self):
                                return s.unlink_cbk (frame, cookie, this,
                                                     op_ret, op_errno,
                                                     preparent, postparent,
                                                     xdata)
                        self.stub_refs["unlink_cbk"] = stub
                        dl.set_unlink_cbk(c_this,stub)
                if "readlink_fop" in funcs:
                        @OpReadlink.fop_type
                        def stub (frame, this, loc, size,
                                  xdata, s=self):
                                return s.readlink_fop (frame, this, loc,
                                                       size, xdata)
                        self.stub_refs["readlink_fop"] = stub
                        dl.set_readlink_fop(c_this,stub)
                if "readlink_cbk" in funcs:
                        @OpReadlink.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  path, buf, xdata, s=self):
                                return s.readlink_cbk (frame, cookie, this,
                                                       op_ret, op_errno,
                                                       path, buf, xdata)
                        self.stub_refs["readlink_cbk"] = stub
                        dl.set_readlink_cbk(c_this,stub)
                if "mkdir_fop" in funcs:
                        @OpMkdir.fop_type
                        def stub (frame, this, loc, mode, umask, xdata,
                                  s=self):
                                return s.mkdir_fop (frame, this, loc, mode,
                                                    umask, xdata)
                        self.stub_refs["mkdir_fop"] = stub
                        dl.set_mkdir_fop(c_this,stub)
                if "mkdir_cbk" in funcs:
                        @OpMkdir.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno, inode,
                                  buf, preparent, postparent, xdata, s=self):
                                return s.mkdir_cbk (frame, cookie, this,
                                                    op_ret, op_errno, inode,
                                                    buf, preparent,
                                                    postparent, xdata)
                        self.stub_refs["mkdir_cbk"] = stub
                        dl.set_mkdir_cbk(c_this,stub)
                if "rmdir_fop" in funcs:
                        @OpRmdir.fop_type
                        def stub (frame, this, loc, xflags,
                                  xdata, s=self):
                                return s.rmdir_fop (frame, this, loc,
                                                    xflags, xdata)
                        self.stub_refs["rmdir_fop"] = stub
                        dl.set_rmdir_fop(c_this,stub)
                if "rmdir_cbk" in funcs:
                        @OpRmdir.cbk_type
                        def stub (frame, cookie, this, op_ret, op_errno,
                                  preparent, postparent, xdata, s=self):
                                return s.rmdir_cbk (frame, cookie, this,
                                                    op_ret, op_errno,
                                                    preparent, postparent,
                                                    xdata)
                        self.stub_refs["rmdir_cbk"] = stub
                        dl.set_rmdir_cbk(c_this,stub)



#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

import os
import sys
import pwd
import time
import fcntl
import shutil
import logging
import socket
from threading import Lock, Thread as baseThread
from errno import EACCES, EAGAIN, EPIPE, ENOTCONN, ECONNABORTED
from errno import EINTR, ENOENT, EPERM, ESTALE, errorcode
from signal import signal, SIGTERM
import select as oselect
from os import waitpid as owaitpid

try:
    from cPickle import PickleError
except ImportError:
    # py 3
    from pickle import PickleError

from gconf import gconf

try:
    # py 3
    from urllib import parse as urllib
except ImportError:
    import urllib

try:
    from hashlib import md5 as md5
except ImportError:
    # py 2.4
    from md5 import new as md5

# auxiliary gfid based access prefix
_CL_AUX_GFID_PFX = ".gfid/"
GF_OP_RETRIES = 20

CHANGELOG_AGENT_SERVER_VERSION = 1.0
CHANGELOG_AGENT_CLIENT_VERSION = 1.0


def escape(s):
    """the chosen flavor of string escaping, used all over
       to turn whatever data to creatable representation"""
    return urllib.quote_plus(s)


def unescape(s):
    """inverse of .escape"""
    return urllib.unquote_plus(s)


def norm(s):
    if s:
        return s.replace('-', '_')


def update_file(path, updater, merger=lambda f: True):
    """update a file in a transaction-like manner"""

    fr = fw = None
    try:
        fd = os.open(path, os.O_CREAT | os.O_RDWR)
        try:
            fr = os.fdopen(fd, 'r+b')
        except:
            os.close(fd)
            raise
        fcntl.lockf(fr, fcntl.LOCK_EX)
        if not merger(fr):
            return

        tmpp = path + '.tmp.' + str(os.getpid())
        fd = os.open(tmpp, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
        try:
            fw = os.fdopen(fd, 'wb', 0)
        except:
            os.close(fd)
            raise
        updater(fw)
        os.fsync(fd)
        os.rename(tmpp, path)
    finally:
        for fx in (fr, fw):
            if fx:
                fx.close()


def create_manifest(fname, content):
    """
    Create manifest file for SSH Control Path
    """
    fd = None
    try:
        fd = os.open(fname, os.O_CREAT | os.O_RDWR)
        try:
            os.write(fd, content)
        except:
            os.close(fd)
            raise
    finally:
        if fd is not None:
            os.close(fd)


def setup_ssh_ctl(ctld, remote_addr, resource_url):
    """
    Setup GConf ssh control path parameters
    """
    gconf.ssh_ctl_dir = ctld
    content = "SLAVE_HOST=%s\nSLAVE_RESOURCE_URL=%s" % (remote_addr,
                                                        resource_url)
    content_md5 = md5hex(content)
    fname = os.path.join(gconf.ssh_ctl_dir,
                         "%s.mft" % content_md5)

    create_manifest(fname, content)
    ssh_ctl_path = os.path.join(gconf.ssh_ctl_dir,
                                "%s.sock" % content_md5)
    gconf.ssh_ctl_args = ["-oControlMaster=auto", "-S", ssh_ctl_path]


def grabfile(fname, content=None):
    """open @fname + contest for its fcntl lock

    @content: if given, set the file content to it
    """
    # damn those messy open() mode codes
    fd = os.open(fname, os.O_CREAT | os.O_RDWR)
    f = os.fdopen(fd, 'r+b', 0)
    try:
        fcntl.lockf(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except:
        ex = sys.exc_info()[1]
        f.close()
        if isinstance(ex, IOError) and ex.errno in (EACCES, EAGAIN):
            # cannot grab, it's taken
            return
        raise
    if content:
        try:
            f.truncate()
            f.write(content)
        except:
            f.close()
            raise
    gconf.permanent_handles.append(f)
    return f


def grabpidfile(fname=None, setpid=True):
    """.grabfile customization for pid files"""
    if not fname:
        fname = gconf.pid_file
    content = None
    if setpid:
        content = str(os.getpid()) + '\n'
    return grabfile(fname, content=content)

final_lock = Lock()


def finalize(*a, **kw):
    """all those messy final steps we go trough upon termination

    Do away with pidfile, ssh control dir and logging.
    """
    final_lock.acquire()
    if getattr(gconf, 'pid_file', None):
        rm_pidf = gconf.pid_file_owned
        if gconf.cpid:
            # exit path from parent branch of daemonization
            rm_pidf = False
            while True:
                f = grabpidfile(setpid=False)
                if not f:
                    # child has already taken over pidfile
                    break
                if os.waitpid(gconf.cpid, os.WNOHANG)[0] == gconf.cpid:
                    # child has terminated
                    rm_pidf = True
                    break
                time.sleep(0.1)
        if rm_pidf:
            try:
                os.unlink(gconf.pid_file)
            except:
                ex = sys.exc_info()[1]
                if ex.errno == ENOENT:
                    pass
                else:
                    raise
    if gconf.ssh_ctl_dir and not gconf.cpid:
        def handle_rm_error(func, path, exc_info):
            if exc_info[1].errno == ENOENT:
                return
            raise exc_info[1]

        shutil.rmtree(gconf.ssh_ctl_dir, onerror=handle_rm_error)
    if getattr(gconf, 'state_socket', None):
        try:
            os.unlink(gconf.state_socket)
        except:
            if sys.exc_info()[0] == OSError:
                pass

    if gconf.log_exit:
        logging.info("exiting.")
    sys.stdout.flush()
    sys.stderr.flush()
    os._exit(kw.get('exval', 0))


def log_raise_exception(excont):
    """top-level exception handler

    Try to some fancy things to cover up we face with an error.
    Translate some weird sounding but well understood exceptions
    into human-friendly lingo
    """
    is_filelog = False
    for h in logging.getLogger().handlers:
        fno = getattr(getattr(h, 'stream', None), 'fileno', None)
        if fno and not os.isatty(fno()):
            is_filelog = True

    exc = sys.exc_info()[1]
    if isinstance(exc, SystemExit):
        excont.exval = exc.code or 0
        raise
    else:
        logtag = None
        if isinstance(exc, GsyncdError):
            if is_filelog:
                logging.error(exc.args[0])
            sys.stderr.write('failure: ' + exc.args[0] + '\n')
        elif isinstance(exc, PickleError) or isinstance(exc, EOFError) or \
            ((isinstance(exc, OSError) or isinstance(exc, IOError)) and
             exc.errno == EPIPE):
            logging.error('connection to peer is broken')
            if hasattr(gconf, 'transport'):
                gconf.transport.wait()
                if gconf.transport.returncode == 127:
                    logging.warn("!!!!!!!!!!!!!")
                    logging.warn('!!! getting "No such file or directory" '
                                 "errors is most likely due to "
                                 "MISCONFIGURATION"
                                 ", please consult https://access.redhat.com"
                                 "/site/documentation/en-US/Red_Hat_Storage"
                                 "/2.1/html/Administration_Guide"
                                 "/chap-User_Guide-Geo_Rep-Preparation-"
                                 "Settingup_Environment.html")
                    logging.warn("!!!!!!!!!!!!!")
                gconf.transport.terminate_geterr()
        elif isinstance(exc, OSError) and exc.errno in (ENOTCONN,
                                                        ECONNABORTED):
            logging.error('glusterfs session went down [%s]',
                          errorcode[exc.errno])
        else:
            logtag = "FAIL"
        if not logtag and logging.getLogger().isEnabledFor(logging.DEBUG):
            logtag = "FULL EXCEPTION TRACE"
        if logtag:
            logging.exception(logtag + ": ")
            sys.stderr.write("failed with %s.\n" % type(exc).__name__)
        excont.exval = 1
        sys.exit(excont.exval)


class FreeObject(object):

    """wildcard class for which any attribute can be set"""

    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)


class Thread(baseThread):

    """thread class flavor for gsyncd

    - always a daemon thread
    - force exit for whole program if thread
      function coughs up an exception
    """

    def __init__(self, *a, **kw):
        tf = kw.get('target')
        if tf:
            def twrap(*aa):
                excont = FreeObject(exval=0)
                try:
                    tf(*aa)
                except:
                    try:
                        log_raise_exception(excont)
                    finally:
                        finalize(exval=excont.exval)
            kw['target'] = twrap
        baseThread.__init__(self, *a, **kw)
        self.setDaemon(True)


class GsyncdError(Exception):
    pass


def getusername(uid=None):
    if uid is None:
        uid = os.geteuid()
    return pwd.getpwuid(uid).pw_name


def privileged():
    return os.geteuid() == 0


def boolify(s):
    """
    Generic string to boolean converter

    return
    - Quick return if string 's' is of type bool
    - True if it's in true_list
    - False if it's in false_list
    - Warn if it's not present in either and return False
    """
    true_list = ['true', 'yes', '1', 'on']
    false_list = ['false', 'no', '0', 'off']

    if isinstance(s, bool):
        return s

    rv = False
    lstr = s.lower()
    if lstr in true_list:
        rv = True
    elif not lstr in false_list:
        logging.warn("Unknown string (%s) in string to boolean conversion "
                     "defaulting to False\n" % (s))

    return rv


def eintr_wrap(func, exc, *a):
    """
    wrapper around syscalls resilient to interrupt caused
    by signals
    """
    while True:
        try:
            return func(*a)
        except exc:
            ex = sys.exc_info()[1]
            if not ex.args[0] == EINTR:
                raise


def select(*a):
    return eintr_wrap(oselect.select, oselect.error, *a)


def waitpid(*a):
    return eintr_wrap(owaitpid, OSError, *a)


def set_term_handler(hook=lambda *a: finalize(*a, **{'exval': 1})):
    signal(SIGTERM, hook)


def is_host_local(host):
    locaddr = False
    for ai in socket.getaddrinfo(host, None):
        # cf. http://github.com/gluster/glusterfs/blob/ce111f47/xlators
        # /mgmt/glusterd/src/glusterd-utils.c#L125
        if ai[0] == socket.AF_INET:
            if ai[-1][0].split(".")[0] == "127":
                locaddr = True
                break
        elif ai[0] == socket.AF_INET6:
            if ai[-1][0] == "::1":
                locaddr = True
                break
        else:
            continue
        try:
            # use ICMP socket to avoid net.ipv4.ip_nonlocal_bind issue,
            # cf. https://bugzilla.redhat.com/show_bug.cgi?id=890587
            s = socket.socket(ai[0], socket.SOCK_RAW, socket.IPPROTO_ICMP)
        except socket.error:
            ex = sys.exc_info()[1]
            if ex.errno != EPERM:
                raise
            f = None
            try:
                f = open("/proc/sys/net/ipv4/ip_nonlocal_bind")
                if int(f.read()) != 0:
                    raise GsyncdError(
                        "non-local bind is set and not allowed to create "
                        "raw sockets, cannot determine if %s is local" % host)
                s = socket.socket(ai[0], socket.SOCK_DGRAM)
            finally:
                if f:
                    f.close()
        try:
            s.bind(ai[-1])
            locaddr = True
            break
        except:
            pass
        s.close()
    return locaddr


def funcode(f):
    fc = getattr(f, 'func_code', None)
    if not fc:
        # python 3
        fc = f.__code__
    return fc


def memoize(f):
    fc = funcode(f)
    fn = fc.co_name

    def ff(self, *a, **kw):
        rv = getattr(self, '_' + fn, None)
        if rv is None:
            rv = f(self, *a, **kw)
            setattr(self, '_' + fn, rv)
        return rv
    return ff


def umask():
    return os.umask(0)


def entry2pb(e):
    return e.rsplit('/', 1)


def gauxpfx():
    return _CL_AUX_GFID_PFX


def md5hex(s):
    return md5(s).hexdigest()


def selfkill(sig=SIGTERM):
    os.kill(os.getpid(), sig)


def errno_wrap(call, arg=[], errnos=[], retry_errnos=[]):
    """ wrapper around calls resilient to errnos.
    """
    nr_tries = 0
    while True:
        try:
            return call(*arg)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno in errnos:
                return ex.errno
            if not ex.errno in retry_errnos:
                raise
            nr_tries += 1
            if nr_tries == GF_OP_RETRIES:
                # probably a screwed state, cannot do much...
                logging.warn('reached maximum retries (%s)...%s' %
                             (repr(arg), ex))
                return ex.errno
            time.sleep(0.250)  # retry the call


def lstat(e):
    try:
        return os.lstat(e)
    except (IOError, OSError):
        ex = sys.exc_info()[1]
        if ex.errno == ENOENT:
            return ex.errno
        else:
            raise


class NoPurgeTimeAvailable(Exception):
    pass


class PartialHistoryAvailable(Exception):
    pass


class ChangelogException(OSError):
    pass



#!/usr/bin/env python

# Copyright (c) 2015 Red Hat, Inc. <http://www.redhat.com/>
# This file is part of GlusterFS.
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

import sys
from errno import ENOENT
import time
from multiprocessing import Process
import os
import xml.etree.cElementTree as etree
from argparse import ArgumentParser, RawDescriptionHelpFormatter, Action
import logging
import shutil

from utils import execute, is_host_local, mkdirp, fail
from utils import setup_logger, human_time, handle_rm_error
from utils import get_changelog_rollover_time, cache_output, create_file
import conf
from changelogdata import OutputMerger

PROG_DESCRIPTION = """
GlusterFS Incremental API
"""
ParseError = etree.ParseError if hasattr(etree, 'ParseError') else SyntaxError

logger = logging.getLogger()
node_outfiles = []


class StoreAbsPath(Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super(StoreAbsPath, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, os.path.abspath(values))


def get_pem_key_path(session, volume):
    return os.path.join(conf.get_opt("session_dir"),
                        session,
                        volume,
                        "%s_%s_secret.pem" % (session, volume))


def node_cmd(host, host_uuid, task, cmd, args, opts):
    """
    Runs command via ssh if host is not local
    """
    localdir = is_host_local(host_uuid)

    pem_key_path = get_pem_key_path(args.session, args.volume)

    if not localdir:
        # prefix with ssh command if not local node
        cmd = ["ssh",
               "-i", pem_key_path,
               "root@%s" % host] + cmd

    execute(cmd, exit_msg="%s - %s failed" % (host, task), logger=logger)

    if opts.get("copy_outfile", False):
        cmd_copy = ["scp",
                    "-i", pem_key_path,
                    "root@%s:/%s" % (host, opts.get("node_outfile")),
                    os.path.dirname(opts.get("node_outfile"))]
        execute(cmd_copy, exit_msg="%s - Copy command failed" % host,
                logger=logger)


def run_cmd_nodes(task, args, **kwargs):
    global node_outfiles
    nodes = get_nodes(args.volume)
    pool = []
    for num, node in enumerate(nodes):
        host, brick = node[1].split(":")
        host_uuid = node[0]
        cmd = []
        opts = {}
        node_outfile = os.path.join(conf.get_opt("working_dir"),
                                    args.session, args.volume,
                                    "tmp_output_%s" % num)

        if task == "pre":
            # If Full backup is requested or start time is zero, use brickfind
            change_detector = conf.get_change_detector("changelog")
            if args.full:
                change_detector = conf.get_change_detector("brickfind")

            node_outfiles.append(node_outfile)

            cmd = [change_detector,
                   args.session,
                   args.volume,
                   brick,
                   node_outfile,
                   str(kwargs.get("start")),
                   "--output-prefix",
                   args.output_prefix] + \
                (["--debug"] if args.debug else []) + \
                (["--only-namespace-changes"] if args.only_namespace_changes
                 else [])

            opts["node_outfile"] = node_outfile
            opts["copy_outfile"] = True
        elif task == "cleanup":
            # After pre run, cleanup the working directory and other temp files
            # Remove the copied node_outfile in main node
            try:
                os.remove(node_outfile)
            except (OSError, IOError):
                logger.warn("Failed to cleanup temporary file %s" %
                            node_outfile)
                pass

            cmd = [conf.get_opt("nodeagent"),
                   "cleanup",
                   args.session,
                   args.volume] + (["--debug"] if args.debug else [])
        elif task == "create":
            # When glusterfind create, create session directory in
            # each brick nodes
            cmd = [conf.get_opt("nodeagent"),
                   "create",
                   args.session,
                   args.volume,
                   brick,
                   kwargs.get("time_to_update")] + \
                (["--debug"] if args.debug else []) + \
                (["--reset-session-time"] if args.reset_session_time
                 else [])
        elif task == "post":
            # Rename pre status file to actual status file in each node
            cmd = [conf.get_opt("nodeagent"),
                   "post",
                   args.session,
                   args.volume,
                   brick] + \
                (["--debug"] if args.debug else [])
        elif task == "delete":
            # When glusterfind delete, cleanup all the session files/dirs
            # from each node.
            cmd = [conf.get_opt("nodeagent"),
                   "delete",
                   args.session,
                   args.volume] + \
                (["--debug"] if args.debug else [])

        if cmd:
            p = Process(target=node_cmd,
                        args=(host, host_uuid, task, cmd, args, opts))
            p.start()
            pool.append(p)

    for num, p in enumerate(pool):
        p.join()
        if p.exitcode != 0:
            logger.warn("Command %s failed in %s" % (task, nodes[num][1]))
            if task in ["create", "delete"]:
                fail("Command %s failed in %s" % (task, nodes[num][1]))
            elif task == "pre" and args.disable_partial:
                sys.exit(1)


@cache_output
def get_nodes(volume):
    """
    Get the gluster volume info xml output and parse to get
    the brick details.
    """
    cmd = ["gluster", 'volume', 'info', volume, "--xml"]
    _, data, _ = execute(cmd,
                         exit_msg="Failed to Run Gluster Volume Info",
                         logger=logger)
    tree = etree.fromstring(data)

    nodes = []
    volume_el = tree.find('volInfo/volumes/volume')
    try:
        for b in volume_el.findall('bricks/brick'):
            nodes.append((b.find('hostUuid').text,
                          b.find('name').text))
    except (ParseError, AttributeError, ValueError) as e:
        fail("Failed to parse Volume Info: %s" % e, logger=logger)

    return nodes


def _get_args():
    parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter,
                            description=PROG_DESCRIPTION)
    subparsers = parser.add_subparsers(dest="mode")

    # create <SESSION> <VOLUME> [--debug] [--force]
    parser_create = subparsers.add_parser('create')
    parser_create.add_argument("session", help="Session Name")
    parser_create.add_argument("volume", help="Volume Name")
    parser_create.add_argument("--debug", help="Debug", action="store_true")
    parser_create.add_argument("--force", help="Force option to recreate "
                               "the session", action="store_true")
    parser_create.add_argument("--reset-session-time",
                               help="Reset Session Time to Current Time",
                               action="store_true")

    # delete <SESSION> <VOLUME> [--debug]
    parser_delete = subparsers.add_parser('delete')
    parser_delete.add_argument("session", help="Session Name")
    parser_delete.add_argument("volume", help="Volume Name")
    parser_delete.add_argument("--debug", help="Debug", action="store_true")

    # list [--session <SESSION>] [--volume <VOLUME>]
    parser_list = subparsers.add_parser('list')
    parser_list.add_argument("--session", help="Session Name", default="")
    parser_list.add_argument("--volume", help="Volume Name", default="")
    parser_list.add_argument("--debug", help="Debug", action="store_true")

    # pre <SESSION> <VOLUME> <OUTFILE>
    #     [--output-prefix <OUTPUT_PREFIX>] [--full]
    parser_pre = subparsers.add_parser('pre')
    parser_pre.add_argument("session", help="Session Name")
    parser_pre.add_argument("volume", help="Volume Name")
    parser_pre.add_argument("outfile", help="Output File", action=StoreAbsPath)
    parser_pre.add_argument("--debug", help="Debug", action="store_true")
    parser_pre.add_argument("--full", help="Full find", action="store_true")
    parser_pre.add_argument("--disable-partial", help="Disable Partial find, "
                            "Fail when one node fails", action="store_true")
    parser_pre.add_argument("--output-prefix", help="File prefix in output",
                            default=".")
    parser_pre.add_argument("--regenerate-outfile",
                            help="Regenerate outfile, discard the outfile "
                            "generated from last pre command",
                            action="store_true")
    parser_pre.add_argument("-N", "--only-namespace-changes",
                            help="List only namespace changes",
                            action="store_true")

    # post <SESSION> <VOLUME>
    parser_post = subparsers.add_parser('post')
    parser_post.add_argument("session", help="Session Name")
    parser_post.add_argument("volume", help="Volume Name")
    parser_post.add_argument("--debug", help="Debug", action="store_true")

    return parser.parse_args()


def ssh_setup(args):
    pem_key_path = get_pem_key_path(args.session, args.volume)

    if not os.path.exists(pem_key_path):
        # Generate ssh-key
        cmd = ["ssh-keygen",
               "-N",
               "",
               "-f",
               pem_key_path]
        execute(cmd,
                exit_msg="Unable to generate ssh key %s"
                % pem_key_path,
                logger=logger)

        logger.info("Ssh key generated %s" % pem_key_path)

    try:
        shutil.copyfile(pem_key_path + ".pub",
                        os.path.join(conf.get_opt("session_dir"),
                                     ".keys",
                                     "%s_%s_secret.pem.pub" % (args.session,
                                                               args.volume)))
    except (IOError, OSError) as e:
        fail("Failed to copy public key to %s: %s"
             % (os.path.join(conf.get_opt("session_dir"), ".keys"), e),
             logger=logger)

    # Copy pub file to all nodes
    cmd = ["gluster",
           "system::",
           "copy",
           "file",
           "/glusterfind/.keys/%s.pub" % os.path.basename(pem_key_path)]

    execute(cmd, exit_msg="Failed to distribute ssh keys", logger=logger)

    logger.info("Distributed ssh key to all nodes of Volume")

    # Add to authorized_keys file in each node
    cmd = ["gluster",
           "system::",
           "execute",
           "add_secret_pub",
           "root",
           "/glusterfind/.keys/%s.pub" % os.path.basename(pem_key_path)]
    execute(cmd,
            exit_msg="Failed to add ssh keys to authorized_keys file",
            logger=logger)

    logger.info("Ssh key added to authorized_keys of Volume nodes")


def mode_create(session_dir, args):
    logger.debug("Init is called - Session: %s, Volume: %s"
                 % (args.session, args.volume))

    execute(["gluster", "volume", "info", args.volume],
            exit_msg="Unable to get volume details",
            logger=logger)

    mkdirp(session_dir, exit_on_err=True, logger=logger)
    mkdirp(os.path.join(session_dir, args.volume), exit_on_err=True,
           logger=logger)
    status_file = os.path.join(session_dir, args.volume, "status")

    if os.path.exists(status_file) and not args.force:
        fail("Session %s already created" % args.session, logger=logger)

    if not os.path.exists(status_file) or args.force:
        ssh_setup(args)

        execute(["gluster", "volume", "set",
                 args.volume, "build-pgfid", "on"],
                exit_msg="Failed to set volume option build-pgfid on",
                logger=logger)
        logger.info("Volume option set %s, build-pgfid on" % args.volume)

        execute(["gluster", "volume", "set",
                 args.volume, "changelog.changelog", "on"],
                exit_msg="Failed to set volume option "
                "changelog.changelog on", logger=logger)
        logger.info("Volume option set %s, changelog.changelog on"
                    % args.volume)

    # Add Rollover time to current time to make sure changelogs
    # will be available if we use this time as start time
    time_to_update = int(time.time()) + get_changelog_rollover_time(
        args.volume)

    run_cmd_nodes("create", args, time_to_update=str(time_to_update))

    if not os.path.exists(status_file) or args.reset_session_time:
        with open(status_file, "w", buffering=0) as f:
            f.write(str(time_to_update))

    sys.exit(0)


def mode_pre(session_dir, args):
    """
    Read from Session file and write to session.pre file
    """
    endtime_to_update = int(time.time()) - get_changelog_rollover_time(
        args.volume)
    status_file = os.path.join(session_dir, args.volume, "status")
    status_file_pre = status_file + ".pre"

    mkdirp(os.path.dirname(args.outfile), exit_on_err=True, logger=logger)

    # If Pre status file exists and running pre command again
    if os.path.exists(status_file_pre) and not args.regenerate_outfile:
        fail("Post command is not run after last pre, "
             "use --regenerate-outfile")

    start = 0
    try:
        with open(status_file) as f:
            start = int(f.read().strip())
    except ValueError:
        pass
    except (OSError, IOError) as e:
        fail("Error Opening Session file %s: %s"
             % (status_file, e), logger=logger)

    logger.debug("Pre is called - Session: %s, Volume: %s, "
                 "Start time: %s, End time: %s"
                 % (args.session, args.volume, start, endtime_to_update))

    run_cmd_nodes("pre", args, start=start)

    # Merger
    if args.full:
        cmd = ["sort", "-u"] + node_outfiles + ["-o", args.outfile]
        execute(cmd,
                exit_msg="Failed to merge output files "
                "collected from nodes", logger=logger)
    else:
        # Read each Changelogs db and generate finaldb
        create_file(args.outfile, exit_on_err=True, logger=logger)
        outfilemerger = OutputMerger(args.outfile + ".db", node_outfiles)

        with open(args.outfile, "a") as f:
            for row in outfilemerger.get():
                # Multiple paths in case of Hardlinks
                paths = row[1].split(",")
                for p in paths:
                    if p == "":
                        continue
                    f.write("%s %s %s\n" % (row[0], p, row[2]))

    try:
        os.remove(args.outfile + ".db")
    except (IOError, OSError):
        pass

    run_cmd_nodes("cleanup", args)

    with open(status_file_pre, "w", buffering=0) as f:
        f.write(str(endtime_to_update))

    sys.stdout.write("Generated output file %s\n" % args.outfile)


def mode_post(session_dir, args):
    """
    If pre session file exists, overwrite session file
    If pre session file does not exists, return ERROR
    """
    status_file = os.path.join(session_dir, args.volume, "status")
    logger.debug("Post is called - Session: %s, Volume: %s"
                 % (args.session, args.volume))
    status_file_pre = status_file + ".pre"

    if os.path.exists(status_file_pre):
        run_cmd_nodes("post", args)
        os.rename(status_file_pre, status_file)
        sys.exit(0)
    else:
        fail("Pre script is not run", logger=logger)


def mode_delete(session_dir, args):
    run_cmd_nodes("delete", args)
    shutil.rmtree(os.path.join(session_dir, args.volume),
                  onerror=handle_rm_error)


def mode_list(session_dir, args):
    """
    List available sessions to stdout, if session name is set
    only list that session.
    """
    if args.session:
        if not os.path.exists(os.path.join(session_dir, args.session)):
            fail("Invalid Session", logger=logger)
        sessions = [args.session]
    else:
        sessions = []
        for d in os.listdir(session_dir):
            if d != ".keys":
                sessions.append(d)

    output = []
    for session in sessions:
        # Session Volume Last Processed
        volnames = os.listdir(os.path.join(session_dir, session))

        for volname in volnames:
            if args.volume and args.volume != volname:
                continue

            status_file = os.path.join(session_dir, session, volname, "status")
            last_processed = None
            try:
                with open(status_file) as f:
                    last_processed = f.read().strip()
            except (OSError, IOError) as e:
                if e.errno == ENOENT:
                    pass
                else:
                    raise
            output.append((session, volname, last_processed))

    if output:
        sys.stdout.write("%s %s %s\n" % ("SESSION".ljust(25),
                                         "VOLUME".ljust(25),
                                         "SESSION TIME".ljust(25)))
        sys.stdout.write("-"*75)
        sys.stdout.write("\n")
    for session, volname, last_processed in output:
        sess_time = 'Session Corrupted'
        if last_processed:
            try:
                sess_time = human_time(last_processed)
            except TypeError:
                sess_time = 'Session Corrupted'
        sys.stdout.write("%s %s %s\n" % (session.ljust(25),
                                         volname.ljust(25),
                                         sess_time.ljust(25)))

    if not output and (args.session or args.volume):
        fail("Invalid Session", logger=logger)


def main():
    args = _get_args()
    mkdirp(conf.get_opt("session_dir"), exit_on_err=True)

    if args.mode == "list":
        session_dir = conf.get_opt("session_dir")
    else:
        session_dir = os.path.join(conf.get_opt("session_dir"),
                                   args.session)

    if not os.path.exists(session_dir) and args.mode not in ["create", "list"]:
        fail("Invalid session %s" % args.session)

    mkdirp(os.path.join(conf.get_opt("log_dir"), args.session, args.volume),
           exit_on_err=True)
    log_file = os.path.join(conf.get_opt("log_dir"),
                            args.session,
                            args.volume,
                            "cli.log")
    setup_logger(logger, log_file, args.debug)

    # globals() will have all the functions already defined.
    # mode_<args.mode> will be the function name to be called
    globals()["mode_" + args.mode](session_dir, args)



#!/usr/bin/python

# Most of this script was written by M S Vishwanath Bhat (vbhat@redhat.com)

import re
import os
import sys
import xattr
import tempfile


def parse_url(url):
    match = re.search(r'([\w - _ @ \.]+)::([\w - _ @ \.]+)', url)
    if match:
        node = match.group(1)
        vol = match.group(2)
    else:
        print 'given url is not a valid url.'
        sys.exit(1)
    return node, vol


def cleanup(master_mnt, slave_mnt):
    try:
        os.system("umount %s" % (master_mnt))
    except:
        print("Failed to unmount the master volume")
    try:
        os.system("umount %s" % (slave_mnt))
    except:
        print("Failed to unmount the slave volume")

    os.removedirs(master_mnt)
    os.removedirs(slave_mnt)


def main():

    masterurl = sys.argv[1]
    slaveurl = sys.argv[2]
    slave_node, slavevol = parse_url(slaveurl)
    master_node, mastervol = parse_url(masterurl)

    master_mnt = tempfile.mkdtemp()
    slave_mnt = tempfile.mkdtemp()

    try:
        print "Mounting master volume on a temp mnt_pnt"
        os.system("glusterfs -s %s --volfile-id %s %s" % (master_node,
                                                          mastervol,
                                                          master_mnt))
    except:
        print("Failed to mount the master volume")
        cleanup(master_mnt, slave_mnt)
        sys.exit(1)

    try:
        print "Mounting slave voluem on a temp mnt_pnt"
        os.system("glusterfs -s %s --volfile-id %s %s" % (slave_node, slavevol,
                                                          slave_mnt))
    except:
        print("Failed to mount the master volume")
        cleanup(master_mnt, slave_mnt)
        sys.exit(1)

    slave_file_list = [slave_mnt]
    for top, dirs, files in os.walk(slave_mnt, topdown=False):
        for subdir in dirs:
            slave_file_list.append(os.path.join(top, subdir))
        for file in files:
            slave_file_list.append(os.path.join(top, file))

    # chdir and then get the gfid, so that you don't need to replace
    gfid_attr = 'glusterfs.gfid'
    ret = 0
    for sfile in slave_file_list:
        mfile = sfile.replace(slave_mnt, master_mnt)
        if xattr.getxattr(sfile, gfid_attr, True) != xattr.getxattr(
                mfile, gfid_attr, True):
            print ("gfid of file %s in slave is different from %s" +
                   " in master" % (sfile, mfile))
            ret = 1

    cleanup(master_mnt, slave_mnt)

    sys.exit(ret)


if __name__ == '__main__':
    if len(sys.argv[1:]) < 2:
        print ("Please pass master volume name and slave url as arguments")
        print ("USAGE : python <script> <master-host>::<master-vol> " +
               "<slave-host>::<slave-vol>")
        sys.exit(1)
    main()



#!/usr/bin/python

import ctypes
import ctypes.util

# find_library does not lookup LD_LIBRARY_PATH and may miss the
# function. In that case, retry with less portable but explicit name.
libgfapi = ctypes.util.find_library("gfapi")
if libgfapi == None:
	libgfapi = "libgfapi.so"
api = ctypes.CDLL(libgfapi,mode=ctypes.RTLD_GLOBAL)

api.glfs_ipc.argtypes = [ ctypes.c_void_p, ctypes.c_int ]
api.glfs_ipc.restype = ctypes.c_int

def do_ipc (host, volume):
	fs = api.glfs_new(volume)
	#api.glfs_set_logging(fs,"/dev/stderr",7)
	api.glfs_set_volfile_server(fs,"tcp",host,24007)

	api.glfs_init(fs)
	ret  = api.glfs_ipc(fs,1470369258)
	api.glfs_fini(fs)

	return ret

if __name__ == "__main__":
	import sys

	try:
		res = apply(do_ipc,sys.argv[1:3])
		print res
	except:
		print "IPC failed (volume not started?)"



#!/usr/bin/python

import sys
import os
import re
import tempfile
import subprocess
from multiprocessing import Pool
import time
from optparse import OptionParser

slave_dict = {}
master_res = ''


def get_arequal_checksum(me, mnt):
    global slave_dict
    master_cmd = ['./tests/utils/arequal-checksum', '-p', mnt]
    print "Calculating  "+me+" checksum ..."
    print ""
    p = subprocess.Popen(master_cmd, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    ret = p.wait()
    stdout, stderr = p.communicate()
    if ret:
        print "Failed to get the checksum of " + me + " with following error"
        print stderr
        return 1
    else:
        return stdout


def get_file_count(me, mnt):
    global slave_dict
    master_cmd = ['find ' + mnt + ' |wc -l']
    print "Calculating  " + me + " files ..."
    print ""
    p = subprocess.Popen(master_cmd, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE, shell=True)
    ret = p.wait()
    stdout, stderr = p.communicate()
    if ret:
        print "Failed to get the count of files in " + me
        + " with following error"
        print stderr
        return 1
    else:
        return stdout.strip()


def compare_checksum(master_mnt, slave_dict):
    proc = len(slave_dict)+1
    pool = Pool(processes=proc)
    master_res = pool.apply_async(get_arequal_checksum, args=("master",
                                                              master_mnt))
    results = [(slave, pool.apply_async(get_arequal_checksum,
                                        args=(slave_dict[slave]["vol"],
                                              slave_dict[slave]["mnt"])))
               for slave in slave_dict]

    pool.close()
    pool.join()
    for slave, result in results:
        slave_dict[slave]["res"] = result.get()
        # exception:  OSError

    master_res = master_res.get()

    print "arequal-checksum of master is : \n %s" % master_res
    for slave in slave_dict:
        print "arequal-checksum of geo_rep_slave %s: \n %s" % (
            slave_dict[slave]["vol"], slave_dict[slave]["res"])

    master_files, master_total = re.findall('Total[\s]+:\s(\w+)', master_res)
    master_reg_meta, master_reg = re.findall('Regular files[\s]+:\s(\w+)',
                                             master_res)[1:]
    master_dir_meta, master_dir = re.findall('Directories[\s]+:\s(\w+)',
                                             master_res)[1:]

    ret = 0
    for slave in slave_dict:
        slave_dict[slave]["files"], slave_dict[slave]["total"] = re.findall(
            'Total[\s]+:\s(\w+)', slave_dict[slave]["res"])
        slave_dict[slave]["reg_meta"], slave_dict[slave]["reg"] = re.findall(
            'Regular files[\s]+:\s(\w+)', slave_dict[slave]["res"])[1:]
        slave_dict[slave]["dir_meta"], slave_dict[slave]["dir"] = re.findall(
            'Directories[\s]+:\s(\w+)', slave_dict[slave]["res"])[1:]

        if master_reg_meta != slave_dict[slave]["reg_meta"]:
            print ("Meta data checksum for regular files doesn't match " +
                   "between master and  "+slave_dict[slave]["vol"])
            ret = 67

        if master_dir_meta != slave_dict[slave]["dir_meta"]:
            print ("Meta data checksum for directories doesn't match " +
                   "between master and "+slave_dict[slave]["vol"])
            ret = 68

        if master_files != slave_dict[slave]["files"]:
            print ("Failed to sync all the files from master to " +
                   slave_dict[slave]["vol"])
            ret = 1

        if master_total != slave_dict[slave]["total"]:
            if master_reg != slave_dict[slave]["reg"]:
                print ("Checksum for regular files doesn't match " +
                       "between master and "+slave_dict[slave]["vol"])
                ret = 1
            elif master_dir != slave_dict[slave]["dir"]:
                print ("Checksum for directories doesn't match between " +
                       "master and "+slave_dict[slave]["vol"])
                ret = 1
            else:
                print ("Checksum for symlinks or others doesn't match " +
                       "between master and "+slave_dict[slave]["vol"])
                ret = 1

        if ret is 0:
            print ("Successfully synced all the files from master " +
                   "to the "+slave_dict[slave]["vol"])

    return ret


def compare_filecount(master_mnt, slave_dict):
    proc = len(slave_dict)+1
    pool = Pool(processes=proc)

    master_res = pool.apply_async(get_file_count, args=("master", master_mnt))
    results = [(slave, pool.apply_async(get_file_count,
                                        args=(slave_dict[slave]["vol"],
                                              slave_dict[slave]["mnt"])))
               for slave in slave_dict]

    pool.close()
    pool.join()
    for slave, result in results:
        slave_dict[slave]["res"] = result.get()

    master_res = master_res.get()
    ret = 0
    for slave in slave_dict:
        if not master_res == slave_dict[slave]["res"]:
            print ("files count between master and " +
                   slave_dict[slave]["vol"]+" doesn't match yet")
            ret = 1

    return ret


def parse_url(url):
    match = re.search(r'([\w - _ @ \.]+)::([\w - _ @ \.]+)', url)
    if match:
        node = match.group(1)
        vol = match.group(2)
    else:
        print 'given url is not a valid.'
        sys.exit(1)
    return node, vol


def cleanup(master_mnt, slave_dict):
    try:
        os.system("umount %s" % (master_mnt))
    except:
        print("Failed to unmount the master volume")

    for slave in slave_dict:

        try:
            os.system("umount %s" % (slave_dict[slave]["mnt"]))
            os.removedirs(slave_dict[slave]["mnt"])
        except:
            print("Failed to unmount the "+slave+" volume")

    os.removedirs(master_mnt)


def main():

    slaves = args[1:]

    masterurl = args[0]
    master_node, mastervol = parse_url(masterurl)
    master_mnt = tempfile.mkdtemp()

    i = 1
    for slave in slaves:
        slave_dict["slave"+str(i)] = {}
        slave_dict["slave"+str(i)]["node"], slave_dict[
            "slave"+str(i)]["vol"] = parse_url(slave)
        slave_dict["slave"+str(i)]["mnt"] = tempfile.mkdtemp()
        i += 1

    try:
        print ("mounting the master volume on "+master_mnt)
        os.system("glusterfs -s  %s --volfile-id %s %s" % (master_node,
                                                           mastervol,
                                                           master_mnt))
        time.sleep(3)
    except:
        print("Failed to mount the master volume")

    for slave in slave_dict:
        print slave
        print slave_dict[slave]
        try:
            print ("mounting the slave volume on "+slave_dict[slave]['mnt'])
            os.system("glusterfs -s %s --volfile-id %s %s" % (
                slave_dict[slave]["node"], slave_dict[slave]["vol"],
                slave_dict[slave]["mnt"]))
            time.sleep(3)
        except:
            print("Failed to mount the "+slave+" volume")

    res = 0
    if option.check == "arequal":
        res = compare_checksum(master_mnt, slave_dict)
    elif option.check == "find":
        res = compare_filecount(master_mnt, slave_dict)
    else:
        print "wrong options given"

    cleanup(master_mnt, slave_dict)

    sys.exit(res)


if __name__ == '__main__':

    usage = "usage: %prog [option] <master-host>::<master-vol> \
    <slave1-host>::<slave1-vol> . . ."
    parser = OptionParser(usage=usage)
    parser.add_option("-c", dest="check", action="store", type="string",
                      default="arequal",
                      help="size of the files to be used [default: %default]")
    (option, args) = parser.parse_args()
    if not args:
        print "usage: <script> [option] <master-host>::<master-vol>\
         <slave1-host>::<slave1-vol> . . ."
        print ""
        sys.exit(1)

    main()



import sys
from gluster.glupy import *

class xlator (Translator):

    def __init__(self, c_this):
        Translator.__init__(self, c_this)

    def lookup_fop(self, frame, this, loc, xdata):
        print "Python xlator: Hello!"
        dl.wind_lookup(frame, POINTER(xlator_t)(), loc, xdata)
        return 0

    def lookup_cbk(self, frame, cookie, this, op_ret, op_errno, inode, buf,
                   xdata, postparent):
        print "Python xlator: Hello again!"
        dl.unwind_lookup(frame, cookie, this, op_ret, op_errno, inode, buf,
                         xdata, postparent)
        return 0



#!/usr/bin/env python

# Copyright (c) 2015 Red Hat, Inc. <http://www.redhat.com/>
# This file is part of GlusterFS.
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

import os
from ctypes import CDLL, get_errno, create_string_buffer, c_ulong, byref
from ctypes import RTLD_GLOBAL
from ctypes.util import find_library


class ChangelogException(OSError):
    pass


libgfc = CDLL(find_library("gfchangelog"), use_errno=True, mode=RTLD_GLOBAL)


def raise_oserr():
    errn = get_errno()
    raise ChangelogException(errn, os.strerror(errn))


def cl_init():
    ret = libgfc.gf_changelog_init(None)
    if ret == -1:
        raise_oserr()


def cl_register(brick, path, log_file, log_level, retries=0):
    ret = libgfc.gf_changelog_register(brick, path, log_file,
                                       log_level, retries)
    if ret == -1:
        raise_oserr()


def cl_history_scan():
    ret = libgfc.gf_history_changelog_scan()
    if ret == -1:
        raise_oserr()

    return ret


def cl_history_changelog(changelog_path, start, end, num_parallel):
    actual_end = c_ulong()
    ret = libgfc.gf_history_changelog(changelog_path, start, end,
                                      num_parallel,
                                      byref(actual_end))
    if ret == -1:
        raise_oserr()

    return actual_end.value


def cl_history_startfresh():
    ret = libgfc.gf_history_changelog_start_fresh()
    if ret == -1:
        raise_oserr()


def cl_history_getchanges():
    """ remove hardcoding for path name length """
    def clsort(f):
        return f.split('.')[-1]

    changes = []
    buf = create_string_buffer('\0', 4096)

    while True:
        ret = libgfc.gf_history_changelog_next_change(buf, 4096)
        if ret in (0, -1):
            break
        changes.append(buf.raw[:ret - 1])
    if ret == -1:
        raise_oserr()

    return sorted(changes, key=clsort)


def cl_history_done(clfile):
    ret = libgfc.gf_history_changelog_done(clfile)
    if ret == -1:
        raise_oserr()



#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

import os
import sys
import time
import signal
import logging
import uuid
import xml.etree.ElementTree as XET
from subprocess import PIPE
from resource import Popen, FILE, GLUSTER, SSH
from threading import Lock
from errno import EEXIST
import re
import random
from gconf import gconf
from syncdutils import select, waitpid
from syncdutils import set_term_handler, is_host_local, GsyncdError
from syncdutils import escape, Thread, finalize, memoize

from gsyncdstatus import GeorepStatus, set_monitor_status


ParseError = XET.ParseError if hasattr(XET, 'ParseError') else SyntaxError


def get_subvol_num(brick_idx, replica_count, disperse_count):
    subvol_size = disperse_count if disperse_count > 0 else replica_count
    cnt = int((brick_idx + 1) / subvol_size)
    rem = (brick_idx + 1) % subvol_size
    if rem > 0:
        return cnt + 1
    else:
        return cnt


def get_slave_bricks_status(host, vol):
    po = Popen(['gluster', '--xml', '--remote-host=' + host,
                'volume', 'status', vol, "detail"],
               stdout=PIPE, stderr=PIPE)
    vix = po.stdout.read()
    po.wait()
    po.terminate_geterr(fail_on_err=False)
    if po.returncode != 0:
        logging.info("Volume status command failed, unable to get "
                     "list of up nodes of %s, returning empty list: %s" %
                     (vol, po.returncode))
        return []
    vi = XET.fromstring(vix)
    if vi.find('opRet').text != '0':
        logging.info("Unable to get list of up nodes of %s, "
                     "returning empty list: %s" %
                     (vol, vi.find('opErrstr').text))
        return []

    up_hosts = set()

    try:
        for el in vi.findall('volStatus/volumes/volume/node'):
            if el.find('status').text == '1':
                up_hosts.add(el.find('hostname').text)
    except (ParseError, AttributeError, ValueError) as e:
        logging.info("Parsing failed to get list of up nodes of %s, "
                     "returning empty list: %s" % (vol, e))

    return list(up_hosts)


class Volinfo(object):

    def __init__(self, vol, host='localhost', prelude=[]):
        po = Popen(prelude + ['gluster', '--xml', '--remote-host=' + host,
                              'volume', 'info', vol],
                   stdout=PIPE, stderr=PIPE)
        vix = po.stdout.read()
        po.wait()
        po.terminate_geterr()
        vi = XET.fromstring(vix)
        if vi.find('opRet').text != '0':
            if prelude:
                via = '(via %s) ' % prelude.join(' ')
            else:
                via = ' '
            raise GsyncdError('getting volume info of %s%s '
                              'failed with errorcode %s',
                              (vol, via, vi.find('opErrno').text))
        self.tree = vi
        self.volume = vol
        self.host = host

    def get(self, elem):
        return self.tree.findall('.//' + elem)

    @property
    @memoize
    def bricks(self):
        def bparse(b):
            host, dirp = b.text.split(':', 2)
            return {'host': host, 'dir': dirp}
        return [bparse(b) for b in self.get('brick')]

    @property
    @memoize
    def uuid(self):
        ids = self.get('id')
        if len(ids) != 1:
            raise GsyncdError("volume info of %s obtained from %s: "
                              "ambiguous uuid",
                              self.volume, self.host)
        return ids[0].text

    @property
    @memoize
    def replica_count(self):
        return int(self.get('replicaCount')[0].text)

    @property
    @memoize
    def disperse_count(self):
        return int(self.get('disperseCount')[0].text)


class Monitor(object):

    """class which spawns and manages gsyncd workers"""

    ST_INIT = 'Initializing...'
    ST_STARTED = 'Started'
    ST_STABLE = 'Active'
    ST_FAULTY = 'Faulty'
    ST_INCON = 'inconsistent'
    _ST_ORD = [ST_STABLE, ST_INIT, ST_FAULTY, ST_INCON]

    def __init__(self):
        self.lock = Lock()
        self.state = {}
        self.status = {}

    @staticmethod
    def terminate():
        # relax one SIGTERM by setting a handler that sets back
        # standard handler
        set_term_handler(lambda *a: set_term_handler())
        # give a chance to graceful exit
        os.kill(-os.getpid(), signal.SIGTERM)

    def monitor(self, w, argv, cpids, agents, slave_vol, slave_host, master):
        """the monitor loop

        Basic logic is a blantantly simple blunt heuristics:
        if spawned client survives 60 secs, it's considered OK.
        This servers us pretty well as it's not vulneralbe to
        any kind of irregular behavior of the child...

        ... well, except for one: if children is hung up on
        waiting for some event, it can survive aeons, still
        will be defunct. So we tweak the above logic to
        expect the worker to send us a signal within 60 secs
        (in the form of closing its end of a pipe). The worker
        does this when it's done with the setup stage
        ready to enter the service loop (note it's the setup
        stage which is vulnerable to hangs -- the full
        blown worker blows up on EPIPE if the net goes down,
        due to the keep-alive thread)
        """
        if not self.status.get(w[0], None):
            self.status[w[0]] = GeorepStatus(gconf.state_file, w[0])

        set_monitor_status(gconf.state_file, self.ST_STARTED)
        self.status[w[0]].set_worker_status(self.ST_INIT)

        ret = 0

        def nwait(p, o=0):
            p2, r = waitpid(p, o)
            if not p2:
                return
            return r

        def exit_signalled(s):
            """ child teminated due to receipt of SIGUSR1 """
            return (os.WIFSIGNALED(s) and (os.WTERMSIG(s) == signal.SIGUSR1))

        def exit_status(s):
            if os.WIFEXITED(s):
                return os.WEXITSTATUS(s)
            return 1

        conn_timeout = int(gconf.connection_timeout)
        while ret in (0, 1):
            remote_host = w[1]
            # Check the status of the connected slave node
            # If the connected slave node is down then try to connect to
            # different up node.
            m = re.match("(ssh|gluster|file):\/\/(.+)@([^:]+):(.+)",
                         remote_host)
            if m:
                current_slave_host = m.group(3)
                slave_up_hosts = get_slave_bricks_status(
                    slave_host, slave_vol)

                if current_slave_host not in slave_up_hosts:
                    if len(slave_up_hosts) > 0:
                        remote_host = "%s://%s@%s:%s" % (m.group(1),
                                                         m.group(2),
                                                         random.choice(
                                                             slave_up_hosts),
                                                         m.group(4))

            # Spawn the worker and agent in lock to avoid fd leak
            self.lock.acquire()

            logging.info('-' * conn_timeout)
            logging.info('starting gsyncd worker')

            # Couple of pipe pairs for RPC communication b/w
            # worker and changelog agent.

            # read/write end for agent
            (ra, ww) = os.pipe()
            # read/write end for worker
            (rw, wa) = os.pipe()

            # spawn the agent process
            apid = os.fork()
            if apid == 0:
                os.execv(sys.executable, argv + ['--local-path', w[0],
                                                 '--agent',
                                                 '--rpc-fd',
                                                 ','.join([str(ra), str(wa),
                                                           str(rw), str(ww)])])
            pr, pw = os.pipe()
            cpid = os.fork()
            if cpid == 0:
                os.close(pr)
                os.execv(sys.executable, argv + ['--feedback-fd', str(pw),
                                                 '--local-path', w[0],
                                                 '--local-id',
                                                 '.' + escape(w[0]),
                                                 '--rpc-fd',
                                                 ','.join([str(rw), str(ww),
                                                           str(ra), str(wa)]),
                                                 '--subvol-num', str(w[2]),
                                                 '--resource-remote',
                                                 remote_host])

            cpids.add(cpid)
            agents.add(apid)
            os.close(pw)

            # close all RPC pipes in monitor
            os.close(ra)
            os.close(wa)
            os.close(rw)
            os.close(ww)
            self.lock.release()

            t0 = time.time()
            so = select((pr,), (), (), conn_timeout)[0]
            os.close(pr)

            if so:
                ret = nwait(cpid, os.WNOHANG)
                if ret is not None:
                    logging.info("worker(%s) died before establishing "
                                 "connection" % w[0])
                    nwait(apid) #wait for agent
                else:
                    logging.debug("worker(%s) connected" % w[0])
                    while time.time() < t0 + conn_timeout:
                        ret = nwait(cpid, os.WNOHANG)
                        if ret is not None:
                            logging.info("worker(%s) died in startup "
                                         "phase" % w[0])
                            nwait(apid) #wait for agent
                            break
                        time.sleep(1)
            else:
                logging.info("worker(%s) not confirmed in %d sec, "
                             "aborting it" % (w[0], conn_timeout))
                os.kill(cpid, signal.SIGKILL)
                nwait(apid) #wait for agent
                ret = nwait(cpid)
            if ret is None:
                self.status[w[0]].set_worker_status(self.ST_STABLE)
                #If worker dies, agent terminates on EOF.
                #So lets wait for agent first.
                nwait(apid)
                ret = nwait(cpid)
            if exit_signalled(ret):
                ret = 0
            else:
                ret = exit_status(ret)
                if ret in (0, 1):
                    self.status[w[0]].set_worker_status(self.ST_FAULTY)
            time.sleep(10)
        self.status[w[0]].set_worker_status(self.ST_INCON)
        return ret

    def multiplex(self, wspx, suuid, slave_vol, slave_host, master):
        argv = sys.argv[:]
        for o in ('-N', '--no-daemon', '--monitor'):
            while o in argv:
                argv.remove(o)
        argv.extend(('-N', '-p', '', '--slave-id', suuid))
        argv.insert(0, os.path.basename(sys.executable))

        cpids = set()
        agents = set()
        ta = []
        for wx in wspx:
            def wmon(w):
                cpid, _ = self.monitor(w, argv, cpids, agents, slave_vol,
                                       slave_host, master)
                time.sleep(1)
                self.lock.acquire()
                for cpid in cpids:
                    os.kill(cpid, signal.SIGKILL)
                for apid in agents:
                    os.kill(apid, signal.SIGKILL)
                self.lock.release()
                finalize(exval=1)
            t = Thread(target=wmon, args=[wx])
            t.start()
            ta.append(t)
        for t in ta:
            t.join()


def distribute(*resources):
    master, slave = resources
    mvol = Volinfo(master.volume, master.host)
    logging.debug('master bricks: ' + repr(mvol.bricks))
    prelude = []
    si = slave
    slave_host = None
    slave_vol = None

    if isinstance(slave, SSH):
        prelude = gconf.ssh_command.split() + [slave.remote_addr]
        si = slave.inner_rsc
        logging.debug('slave SSH gateway: ' + slave.remote_addr)
    if isinstance(si, FILE):
        sbricks = {'host': 'localhost', 'dir': si.path}
        suuid = uuid.uuid5(uuid.NAMESPACE_URL, slave.get_url(canonical=True))
    elif isinstance(si, GLUSTER):
        svol = Volinfo(si.volume, slave.remote_addr.split('@')[-1])
        sbricks = svol.bricks
        suuid = svol.uuid
        slave_host = slave.remote_addr.split('@')[-1]
        slave_vol = si.volume
    else:
        raise GsyncdError("unknown slave type " + slave.url)
    logging.info('slave bricks: ' + repr(sbricks))
    if isinstance(si, FILE):
        slaves = [slave.url]
    else:
        slavenodes = set(b['host'] for b in sbricks)
        if isinstance(slave, SSH) and not gconf.isolated_slave:
            rap = SSH.parse_ssh_address(slave)
            slaves = ['ssh://' + rap['user'] + '@' + h + ':' + si.url
                      for h in slavenodes]
        else:
            slavevols = [h + ':' + si.volume for h in slavenodes]
            if isinstance(slave, SSH):
                slaves = ['ssh://' + rap.remote_addr + ':' + v
                          for v in slavevols]
            else:
                slaves = slavevols

    workerspex = [(brick['dir'], slaves[idx % len(slaves)],
                  get_subvol_num(idx, mvol.replica_count, mvol.disperse_count))
                  for idx, brick in enumerate(mvol.bricks)
                  if is_host_local(brick['host'])]
    logging.info('worker specs: ' + repr(workerspex))
    return workerspex, suuid, slave_vol, slave_host, master


def monitor(*resources):
    # Check if gsyncd restarted in pause state. If
    # yes, send SIGSTOP to negative of monitor pid
    # to go back to pause state.
    if gconf.pause_on_start:
        os.kill(-os.getpid(), signal.SIGSTOP)

    """oh yeah, actually Monitor is used as singleton, too"""
    return Monitor().multiplex(*distribute(*resources))



#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

import os
from ctypes import CDLL, RTLD_GLOBAL, create_string_buffer, get_errno, byref, c_ulong
from ctypes.util import find_library
from syncdutils import ChangelogException


class Changes(object):
    libgfc = CDLL(find_library("gfchangelog"), mode=RTLD_GLOBAL, use_errno=True)

    @classmethod
    def geterrno(cls):
        return get_errno()

    @classmethod
    def raise_changelog_err(cls):
        errn = cls.geterrno()
        raise ChangelogException(errn, os.strerror(errn))

    @classmethod
    def _get_api(cls, call):
        return getattr(cls.libgfc, call)

    @classmethod
    def cl_init(cls):
        ret = cls._get_api('gf_changelog_init')(None)
        if ret == -1:
            cls.raise_changelog_err()

    @classmethod
    def cl_register(cls, brick, path, log_file, log_level, retries=0):
        ret = cls._get_api('gf_changelog_register')(brick, path,
                                                    log_file,
                                                    log_level, retries)
        if ret == -1:
            cls.raise_changelog_err()

    @classmethod
    def cl_scan(cls):
        ret = cls._get_api('gf_changelog_scan')()
        if ret == -1:
            cls.raise_changelog_err()

    @classmethod
    def cl_startfresh(cls):
        ret = cls._get_api('gf_changelog_start_fresh')()
        if ret == -1:
            cls.raise_changelog_err()

    @classmethod
    def cl_getchanges(cls):
        """ remove hardcoding for path name length """
        def clsort(f):
            return f.split('.')[-1]
        changes = []
        buf = create_string_buffer('\0', 4096)
        call = cls._get_api('gf_changelog_next_change')

        while True:
            ret = call(buf, 4096)
            if ret in (0, -1):
                break
            changes.append(buf.raw[:ret - 1])
        if ret == -1:
            cls.raise_changelog_err()
        # cleanup tracker
        cls.cl_startfresh()
        return sorted(changes, key=clsort)

    @classmethod
    def cl_done(cls, clfile):
        ret = cls._get_api('gf_changelog_done')(clfile)
        if ret == -1:
            cls.raise_changelog_err()

    @classmethod
    def cl_history_scan(cls):
        ret = cls._get_api('gf_history_changelog_scan')()
        if ret == -1:
            cls.raise_changelog_err()

        return ret

    @classmethod
    def cl_history_changelog(cls, changelog_path, start, end, num_parallel):
        actual_end = c_ulong()
        ret = cls._get_api('gf_history_changelog')(changelog_path, start, end,
                                                   num_parallel,
                                                   byref(actual_end))
        if ret == -1:
            cls.raise_changelog_err()

        return (ret, actual_end.value)

    @classmethod
    def cl_history_startfresh(cls):
        ret = cls._get_api('gf_history_changelog_start_fresh')()
        if ret == -1:
            cls.raise_changelog_err()

    @classmethod
    def cl_history_getchanges(cls):
        """ remove hardcoding for path name length """
        def clsort(f):
            return f.split('.')[-1]

        changes = []
        buf = create_string_buffer('\0', 4096)
        call = cls._get_api('gf_history_changelog_next_change')

        while True:
            ret = call(buf, 4096)
            if ret in (0, -1):
                break
            changes.append(buf.raw[:ret - 1])
        if ret == -1:
            cls.raise_changelog_err()

        return sorted(changes, key=clsort)

    @classmethod
    def cl_history_done(cls, clfile):
        ret = cls._get_api('gf_history_changelog_done')(clfile)
        if ret == -1:
            cls.raise_changelog_err()



#!/usr/bin/env python

# Copyright (c) 2015 Red Hat, Inc. <http://www.redhat.com/>
# This file is part of GlusterFS.
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

import os
import sys
import time
import xattr
import logging
from argparse import ArgumentParser, RawDescriptionHelpFormatter
import hashlib
import urllib

import libgfchangelog
from utils import mkdirp, symlink_gfid_to_path
from utils import fail, setup_logger, find
from utils import get_changelog_rollover_time
from changelogdata import ChangelogData
import conf


CHANGELOG_LOG_LEVEL = 9
CHANGELOG_CONN_RETRIES = 5
CHANGELOGAPI_NUM_WORKERS = 3
PROG_DESCRIPTION = """
Changelog Crawler
"""
history_turns = 0
history_turn_time = 0

logger = logging.getLogger()


def output_path_prepare(path, output_prefix):
    """
    If Prefix is set, joins to Path, removes ending slash
    and encodes it.
    """
    if output_prefix != ".":
        path = os.path.join(output_prefix, path)
        if path.endswith("/"):
            path = path[0:len(path)-1]

    return urllib.quote_plus(path)


def pgfid_to_path(brick, changelog_data):
    """
    For all the pgfids in table, converts into path using recursive
    readlink.
    """
    # pgfid1 to path1 in case of CREATE/MKNOD/MKDIR/LINK/SYMLINK
    for row in changelog_data.gfidpath_get_distinct("pgfid1", {"path1": ""}):
        # In case of Data/Metadata only, pgfid1 will not be their
        if row[0] == "":
            continue

        path = symlink_gfid_to_path(brick, row[0])
        path = output_path_prepare(path, args.output_prefix)

        changelog_data.gfidpath_set_path1(path, row[0])

    # pgfid2 to path2 in case of RENAME
    for row in changelog_data.gfidpath_get_distinct("pgfid2",
                                                    {"type": "RENAME",
                                                     "path2": ""}):
        # Only in case of Rename pgfid2 exists
        if row[0] == "":
            continue

        path = symlink_gfid_to_path(brick, row[0])
        if path == "":
            continue

        path = output_path_prepare(path, args.output_prefix)
        changelog_data.gfidpath_set_path2(path, row[0])


def populate_pgfid_and_inodegfid(brick, changelog_data):
    """
    For all the DATA/METADATA modifications GFID,
    If symlink, directly convert to Path using Readlink.
    If not symlink, try to get PGFIDs via xattr query and populate it
    to pgfid table, collect inodes in inodegfid table
    """
    for row in changelog_data.gfidpath_get({"path1": "", "type": "MODIFY"}):
        gfid = row[3].strip()
        p = os.path.join(brick, ".glusterfs", gfid[0:2], gfid[2:4], gfid)
        if os.path.islink(p):
            # It is a Directory if GFID backend path is symlink
            path = symlink_gfid_to_path(brick, gfid)
            if path == "":
                continue

            path = output_path_prepare(path, args.output_prefix)

            changelog_data.gfidpath_update({"path1": path},
                                           {"gfid": row[0]})
        else:
            try:
                # INODE and GFID to inodegfid table
                changelog_data.inodegfid_add(os.stat(p).st_ino, gfid)
                file_xattrs = xattr.list(p)
                for x in file_xattrs:
                    if x.startswith("trusted.pgfid."):
                        # PGFID in pgfid table
                        changelog_data.pgfid_add(x.split(".")[-1])
            except (IOError, OSError):
                # All OS Errors ignored, since failures will be logged
                # in End. All GFIDs present in gfidpath table
                continue


def gfid_to_path_using_pgfid(brick, changelog_data, args):
    """
    For all the pgfids collected, Converts to Path and
    does readdir on those directories and looks up inodegfid
    table for matching inode number.
    """
    populate_pgfid_and_inodegfid(brick, changelog_data)

    # If no GFIDs needs conversion to Path
    if not changelog_data.inodegfid_exists({"converted": 0}):
        return

    def inode_filter(path):
        # Looks in inodegfid table, if exists returns
        # inode number else None
        try:
            st = os.lstat(path)
        except (OSError, IOError):
            st = None

        if st and changelog_data.inodegfid_exists({"inode": st.st_ino}):
            return st.st_ino

        return None

    # Length of brick path, to remove from output path
    brick_path_len = len(brick)

    def output_callback(path, inode):
        # For each path found, encodes it and updates path1
        # Also updates converted flag in inodegfid table as 1
        path = path.strip()
        path = path[brick_path_len+1:]

        path = output_path_prepare(path, args.output_prefix)

        changelog_data.append_path1(path, inode)
        changelog_data.inodegfid_update({"converted": 1}, {"inode": inode})

    ignore_dirs = [os.path.join(brick, dirname)
                   for dirname in
                   conf.get_opt("brick_ignore_dirs").split(",")]

    for row in changelog_data.pgfid_get():
        path = symlink_gfid_to_path(brick, row[0])
        find(os.path.join(brick, path),
             callback_func=output_callback,
             filter_func=inode_filter,
             ignore_dirs=ignore_dirs,
             subdirs_crawl=False)


def gfid_to_path_using_batchfind(brick, changelog_data):
    # If all the GFIDs converted using gfid_to_path_using_pgfid
    if not changelog_data.inodegfid_exists({"converted": 0}):
        return

    def inode_filter(path):
        # Looks in inodegfid table, if exists returns
        # inode number else None
        try:
            st = os.lstat(path)
        except (OSError, IOError):
            st = None

        if st and changelog_data.inodegfid_exists({"inode": st.st_ino}):
            return st.st_ino

        return None

    # Length of brick path, to remove from output path
    brick_path_len = len(brick)

    def output_callback(path, inode):
        # For each path found, encodes it and updates path1
        # Also updates converted flag in inodegfid table as 1
        path = path.strip()
        path = path[brick_path_len+1:]
        path = output_path_prepare(path, args.output_prefix)

        changelog_data.append_path1(path, inode)

    ignore_dirs = [os.path.join(brick, dirname)
                   for dirname in
                   conf.get_opt("brick_ignore_dirs").split(",")]

    # Full Namespace Crawl
    find(brick, callback_func=output_callback,
         filter_func=inode_filter,
         ignore_dirs=ignore_dirs)


def parse_changelog_to_db(changelog_data, filename):
    """
    Parses a Changelog file and populates data in gfidpath table
    """
    with open(filename) as f:
        changelogfile = os.path.basename(filename)
        for line in f:
            data = line.strip().split(" ")
            if data[0] == "E" and data[2] in ["CREATE", "MKNOD", "MKDIR"]:
                # CREATE/MKDIR/MKNOD
                changelog_data.when_create_mknod_mkdir(changelogfile, data)
            elif data[0] in ["D", "M"]:
                # DATA/META
                if not args.only_namespace_changes:
                    changelog_data.when_data_meta(changelogfile, data)
            elif data[0] == "E" and data[2] in ["LINK", "SYMLINK"]:
                # LINK/SYMLINK
                changelog_data.when_link_symlink(changelogfile, data)
            elif data[0] == "E" and data[2] == "RENAME":
                # RENAME
                changelog_data.when_rename(changelogfile, data)
            elif data[0] == "E" and data[2] in ["UNLINK", "RMDIR"]:
                # UNLINK/RMDIR
                changelog_data.when_unlink_rmdir(changelogfile, data)


def get_changes(brick, hash_dir, log_file, start, end, args):
    """
    Makes use of libgfchangelog's history API to get changelogs
    containing changes from start and end time. Further collects
    the modified gfids from the changelogs and writes the list
    of gfid to 'gfid_list' file.
    """
    session_dir = os.path.join(conf.get_opt("session_dir"),
                               args.session)
    status_file = os.path.join(session_dir, args.volume,
                               "%s.status" % urllib.quote_plus(args.brick))

    # Get previous session
    try:
        with open(status_file) as f:
            start = int(f.read().strip())
    except (ValueError, OSError, IOError):
        start = args.start

    try:
        libgfchangelog.cl_init()
        libgfchangelog.cl_register(brick, hash_dir, log_file,
                                   CHANGELOG_LOG_LEVEL, CHANGELOG_CONN_RETRIES)
    except libgfchangelog.ChangelogException as e:
        fail("%s Changelog register failed: %s" % (brick, e), logger=logger)

    # Output files to record GFIDs and GFID to Path failure GFIDs
    changelog_data = ChangelogData(args.outfile)

    # Changelogs path(Hard coded to BRICK/.glusterfs/changelogs
    cl_path = os.path.join(brick, ".glusterfs/changelogs")

    # Fail if History fails for requested Start and End
    try:
        actual_end = libgfchangelog.cl_history_changelog(
            cl_path, start, end, CHANGELOGAPI_NUM_WORKERS)
    except libgfchangelog.ChangelogException as e:
        fail("%s Historical Changelogs not available: %s" % (brick, e),
             logger=logger)

    try:
        # scan followed by getchanges till scan returns zero.
        # history_scan() is blocking call, till it gets the number
        # of changelogs to process. Returns zero when no changelogs
        # to be processed. returns positive value as number of changelogs
        # to be processed, which will be fetched using
        # history_getchanges()
        changes = []
        while libgfchangelog.cl_history_scan() > 0:
            changes += libgfchangelog.cl_history_getchanges()

            for change in changes:
                # Ignore if last processed changelog comes
                # again in list
                if change.endswith(".%s" % start):
                    continue
                parse_changelog_to_db(changelog_data, change)
                libgfchangelog.cl_history_done(change)

            changelog_data.commit()
    except libgfchangelog.ChangelogException as e:
        fail("%s Error during Changelog Crawl: %s" % (brick, e),
             logger=logger)

    # Convert all pgfid available from Changelogs
    pgfid_to_path(brick, changelog_data)
    changelog_data.commit()

    # Convert all GFIDs for which no other additional details available
    gfid_to_path_using_pgfid(brick, changelog_data, args)
    changelog_data.commit()

    # If some GFIDs fail to get converted from previous step,
    # convert using find
    gfid_to_path_using_batchfind(brick, changelog_data)
    changelog_data.commit()

    return actual_end


def changelog_crawl(brick, start, end, args):
    """
    Init function, prepares working dir and calls Changelog query
    """
    if brick.endswith("/"):
        brick = brick[0:len(brick)-1]

    # WORKING_DIR/BRICKHASH/OUTFILE
    working_dir = os.path.dirname(args.outfile)
    brickhash = hashlib.sha1(brick)
    brickhash = str(brickhash.hexdigest())
    working_dir = os.path.join(working_dir, brickhash)

    mkdirp(working_dir, exit_on_err=True, logger=logger)

    log_file = os.path.join(conf.get_opt("log_dir"),
                            args.session,
                            args.volume,
                            "changelog.%s.log" % brickhash)

    logger.info("%s Started Changelog Crawl. Start: %s, End: %s"
                % (brick, start, end))
    return get_changes(brick, working_dir, log_file, start, end, args)


def _get_args():
    parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter,
                            description=PROG_DESCRIPTION)

    parser.add_argument("session", help="Session Name")
    parser.add_argument("volume", help="Volume Name")
    parser.add_argument("brick", help="Brick Name")
    parser.add_argument("outfile", help="Output File")
    parser.add_argument("start", help="Start Time", type=int)
    parser.add_argument("--debug", help="Debug", action="store_true")
    parser.add_argument("--output-prefix", help="File prefix in output",
                        default=".")
    parser.add_argument("-N", "--only-namespace-changes",
                        help="List only namespace changes",
                        action="store_true")

    return parser.parse_args()


if __name__ == "__main__":
    args = _get_args()
    mkdirp(os.path.join(conf.get_opt("log_dir"), args.session, args.volume),
           exit_on_err=True)
    log_file = os.path.join(conf.get_opt("log_dir"),
                            args.session,
                            args.volume,
                            "changelog.log")
    setup_logger(logger, log_file, args.debug)

    session_dir = os.path.join(conf.get_opt("session_dir"), args.session)
    status_file = os.path.join(session_dir, args.volume,
                               "%s.status" % urllib.quote_plus(args.brick))
    status_file_pre = status_file + ".pre"
    mkdirp(os.path.join(session_dir, args.volume), exit_on_err=True,
           logger=logger)

    try:
        with open(status_file) as f:
            start = int(f.read().strip())
    except (ValueError, OSError, IOError):
        start = args.start

    end = int(time.time()) - get_changelog_rollover_time(args.volume)
    logger.info("%s Started Changelog Crawl - Start: %s End: %s" % (args.brick,
                                                                    start,
                                                                    end))
    actual_end = changelog_crawl(args.brick, start, end, args)
    with open(status_file_pre, "w", buffering=0) as f:
        f.write(str(actual_end))

    logger.info("%s Finished Changelog Crawl - End: %s" % (args.brick,
                                                           actual_end))
    sys.exit(0)



#!/usr/bin/env python

# Copyright (c) 2015 Red Hat, Inc. <http://www.redhat.com/>
# This file is part of GlusterFS.
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

import sqlite3
import urllib
import os

from utils import RecordType


class OutputMerger(object):
    """
    Class to merge the output files collected from
    different nodes
    """
    def __init__(self, db_path, all_dbs):
        self.conn = sqlite3.connect(db_path)
        self.cursor = self.conn.cursor()
        self.cursor_reader = self.conn.cursor()
        query = "DROP TABLE IF EXISTS finallist"
        self.cursor.execute(query)

        query = """
        CREATE TABLE finallist(
          id     INTEGER PRIMARY KEY AUTOINCREMENT,
          ts     VARCHAR,
          type   VARCHAR,
          gfid   VARCHAR,
          path1  VARCHAR,
          path2  VARCHAR,
          UNIQUE (type, path1, path2) ON CONFLICT IGNORE
        )
        """
        self.cursor.execute(query)

        # If node database exists, read each db and insert into
        # final table. Ignore if combination of TYPE PATH1 PATH2
        # already exists
        for node_db in all_dbs:
            if os.path.exists(node_db):
                conn = sqlite3.connect(node_db)
                cursor = conn.cursor()
                query = """
                SELECT   ts, type, gfid, path1, path2
                FROM     gfidpath
                WHERE    path1 != ''
                ORDER BY id ASC
                """
                for row in cursor.execute(query):
                    self.add_if_not_exists(row[0], row[1], row[2],
                                           row[3], row[4])

        self.conn.commit()

    def add_if_not_exists(self, ts, ty, gfid, path1, path2=""):
        # Adds record to finallist only if not exists
        query = """
        INSERT INTO finallist(ts, type, gfid, path1, path2)
        VALUES(?, ?, ?, ?, ?)
        """
        self.cursor.execute(query, (ts, ty, gfid, path1, path2))

    def get(self):
        query = """SELECT type, path1, path2 FROM finallist
        ORDER BY ts ASC, id ASC"""
        return self.cursor_reader.execute(query)

    def get_failures(self):
        query = """
        SELECT   gfid
        FROM     finallist
        WHERE path1 = '' OR (path2 = '' AND type = 'RENAME')
        """
        return self.cursor_reader.execute(query)


class ChangelogData(object):
    def __init__(self, dbpath):
        self.conn = sqlite3.connect(dbpath)
        self.cursor = self.conn.cursor()
        self.cursor_reader = self.conn.cursor()
        self._create_table_gfidpath()
        self._create_table_pgfid()
        self._create_table_inodegfid()

    def _create_table_gfidpath(self):
        drop_table = "DROP TABLE IF EXISTS gfidpath"
        self.cursor.execute(drop_table)

        create_table = """
        CREATE TABLE gfidpath(
            id     INTEGER PRIMARY KEY AUTOINCREMENT,
            ts     VARCHAR,
            type   VARCHAR,
            gfid   VARCHAR(40),
            pgfid1 VARCHAR(40),
            bn1    VARCHAR(500),
            pgfid2 VARCHAR(40),
            bn2    VARCHAR(500),
            path1  VARCHAR DEFAULT '',
            path2  VARCHAR DEFAULT ''
        )
        """
        self.cursor.execute(create_table)

    def _create_table_inodegfid(self):
        drop_table = "DROP TABLE IF EXISTS inodegfid"
        self.cursor.execute(drop_table)

        create_table = """
        CREATE TABLE inodegfid(
            inode     INTEGER PRIMARY KEY,
            gfid      VARCHAR(40),
            converted INTEGER DEFAULT 0,
            UNIQUE    (inode, gfid) ON CONFLICT IGNORE
        )
        """
        self.cursor.execute(create_table)

    def _create_table_pgfid(self):
        drop_table = "DROP TABLE IF EXISTS pgfid"
        self.cursor.execute(drop_table)

        create_table = """
        CREATE TABLE pgfid(
            pgfid  VARCHAR(40) PRIMARY KEY,
            UNIQUE (pgfid) ON CONFLICT IGNORE
        )
        """
        self.cursor.execute(create_table)

    def _get(self, tablename, filters):
        # SELECT * FROM <TABLENAME> WHERE <CONDITION>
        params = []
        query = "SELECT * FROM %s WHERE 1=1" % tablename

        for key, value in filters.items():
            query += " AND %s = ?" % key
            params.append(value)

        return self.cursor_reader.execute(query, params)

    def _get_distinct(self, tablename, distinct_field, filters):
        # SELECT DISTINCT <COL> FROM <TABLENAME> WHERE <CONDITION>
        params = []
        query = "SELECT DISTINCT %s FROM %s WHERE 1=1" % (distinct_field,
                                                          tablename)

        for key, value in filters.items():
            query += " AND %s = ?" % key
            params.append(value)

        return self.cursor_reader.execute(query, params)

    def _delete(self, tablename, filters):
        # DELETE FROM <TABLENAME> WHERE <CONDITIONS>
        query = "DELETE FROM %s WHERE 1=1" % tablename
        params = []

        for key, value in filters.items():
            query += " AND %s = ?" % key
            params.append(value)

        self.cursor.execute(query, params)

    def _add(self, tablename, data):
        # INSERT INTO <TABLENAME>(<col1>, <col2>..) VALUES(?,?..)
        query = "INSERT INTO %s(" % tablename
        fields = []
        params = []
        for key, value in data.items():
            fields.append(key)
            params.append(value)

        values_substitute = len(fields)*["?"]
        query += "%s) VALUES(%s)" % (",".join(fields),
                                     ",".join(values_substitute))
        self.cursor.execute(query, params)

    def _update(self, tablename, data, filters):
        # UPDATE <TABLENAME> SET col1 = ?,.. WHERE col1=? AND ..
        params = []
        update_fields = []
        for key, value in data.items():
            update_fields.append("%s = ?" % key)
            params.append(value)

        query = "UPDATE %s SET %s WHERE 1 = 1" % (tablename,
                                                  ", ".join(update_fields))

        for key, value in filters.items():
            query += " AND %s = ?" % key
            params.append(value)

        self.cursor.execute(query, params)

    def _exists(self, tablename, filters):
        if not filters:
            return False

        query = "SELECT COUNT(1) FROM %s WHERE 1=1" % tablename
        params = []

        for key, value in filters.items():
            query += " AND %s = ?" % key
            params.append(value)

        self.cursor.execute(query, params)
        row = self.cursor.fetchone()
        return True if row[0] > 0 else False

    def gfidpath_add(self, changelogfile, ty, gfid, pgfid1="", bn1="",
                     pgfid2="", bn2="", path1="", path2=""):
        self._add("gfidpath", {
            "ts": changelogfile.split(".")[-1],
            "type": ty,
            "gfid": gfid,
            "pgfid1": pgfid1,
            "bn1": bn1,
            "pgfid2": pgfid2,
            "bn2": bn2,
            "path1": path1,
            "path2": path2
        })

    def gfidpath_update(self, data, filters):
        self._update("gfidpath", data, filters)

    def gfidpath_delete(self, filters):
        self._delete("gfidpath", filters)

    def gfidpath_exists(self, filters):
        return self._exists("gfidpath", filters)

    def gfidpath_get(self, filters={}):
        return self._get("gfidpath", filters)

    def gfidpath_get_distinct(self, distinct_field, filters={}):
        return self._get_distinct("gfidpath", distinct_field, filters)

    def pgfid_add(self, pgfid):
        self._add("pgfid", {
            "pgfid": pgfid
        })

    def pgfid_update(self, data, filters):
        self._update("pgfid", data, filters)

    def pgfid_get(self, filters={}):
        return self._get("pgfid", filters)

    def pgfid_get_distinct(self, distinct_field, filters={}):
        return self._get_distinct("pgfid", distinct_field, filters)

    def pgfid_exists(self, filters):
        return self._exists("pgfid", filters)

    def inodegfid_add(self, inode, gfid, converted=0):
        self._add("inodegfid", {
            "inode": inode,
            "gfid": gfid,
            "converted": converted
        })

    def inodegfid_update(self, data, filters):
        self._update("inodegfid", data, filters)

    def inodegfid_get(self, filters={}):
        return self._get("inodegfid", filters)

    def inodegfid_get_distinct(self, distinct_field, filters={}):
        return self._get_distinct("inodegfid", distinct_field, filters)

    def inodegfid_exists(self, filters):
        return self._exists("inodegfid", filters)

    def append_path1(self, path, inode):
        # || is for concatenate in SQL
        query = """UPDATE gfidpath SET path1 = ',' || ?
        WHERE gfid IN (SELECT gfid FROM inodegfid WHERE inode = ?)"""
        self.cursor.execute(query, (path, inode))

    def gfidpath_set_path1(self, path1, pgfid1):
        # || is for concatenate in SQL
        if path1 == "":
            update_str1 = "? || bn1"
            update_str2 = "? || bn2"
        else:
            update_str1 = "? || '%2F' || bn1"
            update_str2 = "? || '%2F' || bn2"

        query = """UPDATE gfidpath SET path1 = %s
        WHERE pgfid1 = ?""" % update_str1
        self.cursor.execute(query, (path1, pgfid1))

        # Set Path2 if pgfid1 and pgfid2 are same
        query = """UPDATE gfidpath SET path2 = %s
        WHERE pgfid2 = ?""" % update_str2
        self.cursor.execute(query, (path1, pgfid1))

    def gfidpath_set_path2(self, path2, pgfid2):
        # || is for concatenate in SQL
        if path2 == "":
            update_str = "? || bn2"
        else:
            update_str = "? || '%2F' || bn2"

        query = """UPDATE gfidpath SET path2 = %s
        WHERE pgfid2 = ?""" % update_str
        self.cursor.execute(query, (path2, pgfid2))

    def when_create_mknod_mkdir(self, changelogfile, data):
        # E <GFID> <MKNOD|CREATE|MKDIR> <MODE> <USER> <GRP> <PGFID>/<BNAME>
        # Add the Entry to DB
        pgfid1, bn1 = urllib.unquote_plus(data[6]).split("/", 1)

        # Quote again the basename
        bn1 = urllib.quote_plus(bn1.strip())

        self.gfidpath_add(changelogfile, RecordType.NEW, data[1], pgfid1, bn1)

    def when_rename(self, changelogfile, data):
        # E <GFID> RENAME <OLD_PGFID>/<BNAME> <PGFID>/<BNAME>
        pgfid1, bn1 = urllib.unquote_plus(data[3]).split("/", 1)
        pgfid2, bn2 = urllib.unquote_plus(data[4]).split("/", 1)

        # Quote again the basename
        bn1 = urllib.quote_plus(bn1.strip())
        bn2 = urllib.quote_plus(bn2.strip())

        if self.gfidpath_exists({"gfid": data[1], "type": "NEW",
                                 "pgfid1": pgfid1, "bn1": bn1}):
            # If <OLD_PGFID>/<BNAME> is same as CREATE, Update
            # <NEW_PGFID>/<BNAME> in NEW.
            self.gfidpath_update({"pgfid1": pgfid2, "bn1": bn2},
                                 {"gfid": data[1], "type": "NEW",
                                  "pgfid1": pgfid1, "bn1": bn1})
        elif self.gfidpath_exists({"gfid": data[1], "type": "RENAME",
                                   "pgfid2": pgfid1, "bn2": bn1}):
            # If <OLD_PGFID>/<BNAME> is same as <PGFID2>/<BN2>(may be previous
            # RENAME) then UPDATE <NEW_PGFID>/<BNAME> as <PGFID2>/<BN2>
            self.gfidpath_update({"pgfid2": pgfid2, "bn2": bn2},
                                 {"gfid": data[1], "type": "RENAME",
                                 "pgfid2": pgfid1, "bn2": bn1})
        else:
            # Else insert as RENAME
            self.gfidpath_add(changelogfile, RecordType.RENAME, data[1],
                              pgfid1, bn1, pgfid2, bn2)

    def when_link_symlink(self, changelogfile, data):
        # E <GFID> <LINK|SYMLINK> <PGFID>/<BASENAME>
        # Add as New record in Db as Type NEW
        pgfid1, bn1 = urllib.unquote_plus(data[3]).split("/", 1)

        # Quote again the basename
        bn1 = urllib.quote_plus(bn1.strip())

        self.gfidpath_add(changelogfile, RecordType.NEW, data[1], pgfid1, bn1)

    def when_data_meta(self, changelogfile, data):
        # If GFID row exists, Ignore else Add to Db
        if not self.gfidpath_exists({"gfid": data[1]}):
            self.gfidpath_add(changelogfile, RecordType.MODIFY, data[1])

    def when_unlink_rmdir(self, changelogfile, data):
        # E <GFID> <UNLINK|RMDIR> <PGFID>/<BASENAME>
        pgfid1, bn1 = urllib.unquote_plus(data[3]).split("/", 1)
        # Quote again the basename
        bn1 = urllib.quote_plus(bn1.strip())
        deleted_path = data[4] if len(data) == 5 else ""

        if self.gfidpath_exists({"gfid": data[1], "type": "NEW",
                                 "pgfid1": pgfid1, "bn1": bn1}):
            # If path exists in table as NEW with same GFID
            # Delete that row
            self.gfidpath_delete({"gfid": data[1], "type": "NEW",
                                  "pgfid1": pgfid1, "bn1": bn1})
        else:
            # Else Record as DELETE
            self.gfidpath_add(changelogfile, RecordType.DELETE, data[1],
                              pgfid1, bn1, path1=deleted_path)

        # Update path1 as deleted_path if pgfid1 and bn1 is same as deleted
        self.gfidpath_update({"path1": deleted_path}, {"gfid": data[1],
                                                       "pgfid1": pgfid1,
                                                       "bn1": bn1})

        # Update path2 as deleted_path if pgfid2 and bn2 is same as deleted
        self.gfidpath_update({"path2": deleted_path}, {
            "type": RecordType.RENAME,
            "gfid": data[1],
            "pgfid2": pgfid1,
            "bn2": bn1})

        # If deleted directory is parent for somebody
        query1 = """UPDATE gfidpath SET path1 = ? || '%2F' || bn1
        WHERE pgfid1 = ? AND path1 != ''"""
        self.cursor.execute(query1, (deleted_path, data[1]))

        query1 = """UPDATE gfidpath SET path2 = ? || '%2F' || bn1
        WHERE pgfid2 = ? AND path2 != ''"""
        self.cursor.execute(query1, (deleted_path, data[1]))

    def commit(self):
        self.conn.commit()
